# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from fractions import Fraction
import itertools
import json
import sys
if __name__ == '__main__':
    sys.path.append('..')
    sys.path.append('.')

from pyzx_param.graph import Graph
from pyzx_param.utils import EdgeType, VertexType
from pyzx_param.generate import identity

import numpy as np
from pyzx_param.tensor import compare_tensors
from pyzx_param.graph.scalar import DyadicNumber, Scalar



class TestGraphBasicMethods(unittest.TestCase):

    def test_empty_graph(self):
        g = Graph()
        self.assertEqual(g.num_vertices(),0)
        self.assertEqual(g.num_edges(),0)

    def test_add_remove_vertices(self):
        g = Graph()
        v = g.add_vertex()
        self.assertEqual(g.num_vertices(),1)
        g.add_vertices(3)
        self.assertEqual(g.num_vertices(),4)
        g.remove_vertex(v)
        self.assertEqual(g.num_vertices(),3)

    def test_edges(self):
        g = Graph()
        v1, v2, v3 = g.add_vertices(3)
        g.add_edge((v1,v2))
        self.assertEqual(g.num_edges(),1)
        self.assertTrue(g.connected(v1,v2))
        self.assertTrue(v2 in g.neighbors(v1))
        self.assertEqual(g.vertex_degree(v1), 1)
        self.assertFalse(g.connected(v1,v3))
        e = g.edge(v1,v2)
        self.assertEqual(g.edge_type(e),EdgeType.SIMPLE)
        g.set_edge_type(e,EdgeType.HADAMARD)
        self.assertEqual(g.edge_type(e),EdgeType.HADAMARD)
        g.remove_edge(e)
        self.assertEqual(g.num_edges(),0)
        self.assertFalse(g.connected(v1,v2))

    def test_self_loop(self):
        g = Graph()
        v1 = g.add_vertex(VertexType.X, 0, 0)
        v2 = g.add_vertex(VertexType.Z, 0, 0)
        g.add_edge((v1,v1))
        g.add_edge((v2,v2))

        self.assertEqual(g.num_edges(),0)
        self.assertEqual(g.vertex_degree(v1), 0)
        self.assertEqual(g.vertex_degree(v2), 0)
        self.assertEqual(g.phases()[0], 0)
        self.assertEqual(g.phases()[1], 0)

    def test_self_loop_had_edge(self):
        g = Graph()
        v1 = g.add_vertex(VertexType.X, 0, 0)
        v2 = g.add_vertex(VertexType.Z, 0, 0)
        g.add_edge((v1,v1), edgetype=EdgeType.HADAMARD)
        g.add_edge((v2,v2), edgetype=EdgeType.HADAMARD)

        self.assertEqual(g.num_edges(),0)
        self.assertEqual(g.vertex_degree(v1), 0)
        self.assertEqual(g.vertex_degree(v2), 0)
        self.assertEqual(g.phases()[0], 1)
        self.assertEqual(g.phases()[1], 1)

    def test_set_attributes(self):
        g = Graph()
        v = g.add_vertex()
        self.assertEqual(g.phase(v),0)
        g.set_phase(v,1)
        self.assertEqual(g.phase(v),1)
        self.assertEqual(g.type(v),VertexType.BOUNDARY)
        g.set_type(v,VertexType.X)
        self.assertEqual(g.type(v),VertexType.X)
        self.assertFalse(g.is_ground(v))
        g.set_ground(v)
        self.assertTrue(g.is_ground(v))
        g.set_row(v,3)
        self.assertEqual(g.row(v),3)
        g.set_qubit(v,2)
        self.assertEqual(g.qubit(v),2)

    def test_add_edge_table_same_type(self):
        g = Graph()
        v1, v2 = g.add_vertices(2)
        g.set_type(v1,VertexType.Z)
        g.set_type(v2,VertexType.Z)
        etab = {(v1,v2):[2,0]}
        g.add_edge_table(etab)
        self.assertTrue(g.connected(v1,v2))
        self.assertEqual((g.phase(v1),g.phase(v2)),(0,0))
        g.remove_edge(g.edge(v1,v2))
        self.assertFalse(g.connected(v1,v2))
        etab = {(v1,v2):[0,2]}
        g.add_edge_table(etab)
        self.assertFalse(g.connected(v1,v2))
        etab = {(v1,v2): [1,1]}
        g.add_edge_table(etab)
        self.assertEqual(g.edge_type(g.edge(v1,v2)),EdgeType.SIMPLE)
        self.assertTrue((g.phase(v1)==1  and g.phase(v2)==0) or (g.phase(v1)==0 and g.phase(v2)==1))

    def test_add_edge_table_different_type(self):
        g = Graph()
        v1, v2 = g.add_vertices(2)
        g.set_type(v1,VertexType.Z)
        g.set_type(v2,VertexType.X)
        etab = {(v1,v2):[0,2]}
        g.add_edge_table(etab)
        self.assertTrue(g.connected(v1,v2))
        self.assertEqual((g.phase(v1),g.phase(v2)),(0,0))
        g.remove_edge(g.edge(v1,v2))
        self.assertFalse(g.connected(v1,v2))
        etab = {(v1,v2):[2,0]}
        g.add_edge_table(etab)
        self.assertFalse(g.connected(v1,v2))
        etab = {(v1,v2): [1,1]}
        g.add_edge_table(etab)
        self.assertEqual(g.edge_type(g.edge(v1,v2)),EdgeType.HADAMARD)
        self.assertTrue((g.phase(v1)==1  and g.phase(v2)==0) or (g.phase(v1)==0 and g.phase(v2)==1))

    def test_copy(self):
        g = Graph()
        v1, v2 = g.add_vertices(2)
        g.add_edge((v1,v2),EdgeType.HADAMARD)
        g2 = g.copy()
        self.assertEqual(g.num_vertices(),g2.num_vertices())
        self.assertEqual(g.num_edges(),g2.num_edges())
        v1, v2 = list(g2.vertices())
        self.assertEqual(g.edge_type(g.edge(v1,v2)),EdgeType.HADAMARD)
        
    def test_adjoint_scalar(self):
        g = Graph()
        scalar = Scalar()
        scalar.phase = Fraction(1, 4)
        scalar.phasenodes = [Fraction(1, 2), Fraction(1, 4)]
        scalar.floatfactor = DyadicNumber(1, 3, 3, 1, 5)
        scalar.power2 = 2
        g.scalar = scalar
        g_adj = g.adjoint()
        self.assertAlmostEqual(g_adj.scalar.to_number(), scalar.to_number().conjugate())

    @unittest.skipUnless(np, "numpy needs to be installed for this to run")
    def test_remove_isolated_vertex_preserves_semantics(self):
        g = Graph()
        v = g.add_vertex(VertexType.Z,0,0)
        g2 = g.copy()
        g2.remove_isolated_vertices()
        self.assertTrue(compare_tensors(g,g2,preserve_scalar=True))
        self.assertAlmostEqual(g2.scalar.to_number(),2)
        g.set_phase(v,Fraction(1))
        g2 = g.copy()
        g2.remove_isolated_vertices()
        self.assertTrue(compare_tensors(g,g2))
        self.assertAlmostEqual(g2.scalar.to_number(),0)

    @unittest.skipUnless(np, "numpy needs to be installed for this to run")
    def test_remove_isolated_pair_preserves_semantics(self):
        for i,j in itertools.product([VertexType.Z,VertexType.X],repeat=2):
            for k in [EdgeType.SIMPLE, EdgeType.HADAMARD]:
                for phase1, phase2 in itertools.product([0,1,2],[0,4,5]):
                    with self.subTest(i=i,j=j,k=k,phase1=phase1,phase2=phase2):
                        g = Graph()
                        v = g.add_vertex(i,0,0,phase=phase1)
                        w = g.add_vertex(j,1,0,phase=phase2)
                        g.add_edge((v,w),k)
                        g2 = g.copy()
                        g2.remove_isolated_vertices()
                        self.assertEqual(g2.num_vertices(),0)
                        self.assertTrue(compare_tensors(g,g2))

class TestGraphCircuitMethods(unittest.TestCase):

    def setUp(self):
        """Sets up a two qubit circuit containing a single CNOT with some phases."""
        self.graph = Graph()
        g = self.graph
        i1 = g.add_vertex(VertexType.BOUNDARY,0,0) #add_vertex(type,qubit_index,row_index,phase=0)
        i2 = g.add_vertex(VertexType.BOUNDARY,1,0)
        g.set_inputs((i1,i2))
        v = g.add_vertex(VertexType.Z,0,1,Fraction(1,2))
        w = g.add_vertex(VertexType.X,1,1,Fraction(1,1))
        o1 = g.add_vertex(VertexType.BOUNDARY,0,2)
        o2 = g.add_vertex(VertexType.BOUNDARY,1,2)
        g.set_outputs((o1, o2))
        g.add_edges([(i1,v),(i2,w),(v,w),(v,o1),(w,o2)])
        self.i1, self.i2, self.v, self.w, self.o1, self.o2 = i1, i2, v, w, o1, o2

    def test_qubit_index_and_depth(self):
        g = self.graph
        self.assertEqual(g.depth(),2)
        self.assertEqual(g.qubit_count(),2)

    def test_adjoint(self):
        g = self.graph
        adj = g.adjoint()
        self.assertEqual(g.num_vertices(),adj.num_vertices())
        self.assertEqual(g.num_edges(),adj.num_edges())
        self.assertEqual(g.depth(), adj.depth())
        self.assertEqual(g.qubit_count(), adj.qubit_count())

        v = [i for i in g.vertices() if g.type(i)==VertexType.Z][0]
        w = [i for i in adj.vertices() if adj.type(i)==VertexType.Z][0]
        self.assertEqual(g.phase(v),(-adj.phase(v))%2)
        self.assertEqual(g.vertex_degree(v),adj.vertex_degree(w))

    def test_compose_basic(self):
        g = self.graph.copy()
        g.compose(g)
        self.assertEqual((g.num_inputs(),g.num_outputs()),(2,2))

    @unittest.skipUnless(np, "numpy needs to be installed for this to run")
    def test_compose_unitary(self):
        g = self.graph
        g.set_edge_type(g.edge(self.v,self.o1),EdgeType.HADAMARD)
        g2 = g.adjoint()
        g2.compose(g)
        self.assertTrue(compare_tensors(g2,identity(2), False))




# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from argparse import ArgumentTypeError
from enum import IntEnum
from fractions import Fraction
from typing import Union, Optional, List, Dict, Any
from typing_extensions import Literal, Final

from .symbolic import Poly

FloatInt = Union[float,int]
FractionLike = Union[Fraction,int,Poly]


class VertexType(IntEnum):
    """Type of a vertex in the graph."""
    BOUNDARY = 0
    Z = 1
    X = 2
    H_BOX = 3
    W_INPUT = 4
    W_OUTPUT = 5
    Z_BOX = 6
    DUMMY = 99

def vertex_is_zx(ty: VertexType) -> bool:
    """Check if a vertex type corresponds to a green or red spider."""
    return ty in (VertexType.Z, VertexType.X)

def toggle_vertex(ty: VertexType) -> VertexType:
    """Swap the X and Z vertex types."""
    if not vertex_is_zx(ty):
        return ty
    return VertexType.Z if ty == VertexType.X else VertexType.X

def vertex_is_z_like(ty: VertexType) -> bool:
    """Check if a vertex type corresponds to a Z spider or Z box."""
    return ty == VertexType.Z or ty == VertexType.Z_BOX

def vertex_is_zx_like(ty: VertexType) -> bool:
    """Check if a vertex type corresponds to a Z or X spider or Z box."""
    return vertex_is_z_like(ty) or ty == VertexType.X

def vertex_is_w(ty: VertexType) -> bool:
    return ty == VertexType.W_INPUT or ty == VertexType.W_OUTPUT

def get_w_partner(g, v):
    assert vertex_is_w(g.type(v))
    for edge in g.incident_edges(v):
        if g.edge_type(edge) == EdgeType.W_IO:
            u = edge[0] if edge[1] == v else edge[1]
            assert vertex_is_w(g.type(u))
            return u
    assert False

def get_w_io(g, v):
    v2 = get_w_partner(g, v)
    if g.type(v) == VertexType.W_INPUT:
        return v, v2
    return v2, v


class EdgeType(IntEnum):
    """Type of an edge in the graph."""
    SIMPLE = 1
    HADAMARD = 2
    W_IO = 3

def toggle_edge(ty: EdgeType) -> EdgeType:
    """Swap the regular and Hadamard edge types."""
    return EdgeType.HADAMARD if ty == EdgeType.SIMPLE else EdgeType.SIMPLE

def phase_to_s(a: FractionLike, t:VertexType=VertexType.Z, poly_with_pi:bool=False) -> str:
    if isinstance(a, Fraction) or isinstance(a, int):
        return phase_fraction_to_s(a, t)
    else: # a is a Poly
        if poly_with_pi:
            return f"({a})\u03c0"
        else:
            return str(a)

def phase_fraction_to_s(a: FractionLike, t:VertexType=VertexType.Z) -> str:
    if (a == 0 and t != VertexType.H_BOX): return ''
    if (a == 1 and t == VertexType.H_BOX): return ''
    if isinstance(a, Poly):
        return str(a)
    a = Fraction(a)

    if a == 0: return '0'
    simstr = ''
    if a.denominator > 256:
        a = a.limit_denominator(256)
        simstr = '~'

    ns = '' if a.numerator == 1 else str(a.numerator)
    ds = '' if a.denominator == 1 else '/' + str(a.denominator)

    # unicode 0x03c0 = pi
    return simstr + ns + '\u03c0' + ds

def phase_is_clifford(phase: FractionLike):
    if isinstance(phase, (Fraction, int)):
        return phase in [Fraction(i, 2) for i in range(4)]
    else:
        return phase.is_clifford

def phase_is_pauli(phase: FractionLike):
    if isinstance(phase, (Fraction, int)):
        return phase in (0, 1)
    else:
        return phase.is_pauli

tikz_classes = {
    'boundary': 'none',
    'Z': 'Z dot',
    'X': 'X dot',
    'Z phase': 'Z phase dot',
    'X phase': 'X phase dot',
    'Z box': 'Z box',
    'H': 'hadamard',
    'W': 'W triangle',
    'W input': 'W input',
    'dummy': 'text',
    'edge': '',
    'H-edge': 'hadamard edge',
    'W-io-edge': 'W io edge'
}

original_colors = {
    'edge': '#000000',
    'Hedge': '#0088ff',
    'Xedge': '#999999',
    'boundary': '#000000',
    'X': '#ff8888',
    'Y': '#aabbff',
    'Z': '#ccffcc',
    'H': '#ffff66',
    'W': '#000000',
    'Zalt': '#ccffcc',
    'Walt': '#000000',
    'Xdark': '#ff8888',
    'Ydark': '#aabbff',
    'Zdark': '#99dd99',
}

rgb_colors = original_colors.copy()
rgb_colors['Y'] = original_colors['Z']
rgb_colors['Z'] = original_colors['Y']
rgb_colors['Ydark'] = original_colors['Zdark']
rgb_colors['Zdark'] = original_colors['Ydark']
rgb_colors['Hedge'] = '#ff6600'

grayscale_colors = {
    'edge': '#000000',
    'Hedge': '#888888',
    'Xedge': '#dddddd',
    'boundary': '#000000',
    'X': '#666666',
    'Y': '#9999dd',
    'Z': '#dddddd',
    'H': '#eeeeee',
    'W': '#000000',
    'Zalt': '#dddddd',
    'Walt': '#000000',
    'Xdark': '#666666',
    'Ydark': '#9999dd',
    'Zdark': '#dddddd',
}

class Settings(object): # namespace class
    mode: Literal["notebook", "browser", "shell", ""] = "shell"
    drawing_backend: Literal["d3","matplotlib"] = "d3"
    drawing_auto_hbox: bool = False
    javascript_location: str = "" # Path to javascript files of pyzx
    d3_load_string: str = ""
    tikzit_location: str = "" # Path to tikzit executable
    quantomatic_location: str = "" # Path to quantomatic executable
    topt_command: Optional[List[str]] = None # Argument-separated command to run TOpt such as ["wsl", "./TOpt"]
    show_labels: bool = False
    tikz_classes: Dict[str,str] = tikz_classes
    default_qasm_version: int = 2
    colors: Dict[str, str] = original_colors
    javascript_importmap: Dict[str, Any] = {
        "imports": {
            "three": "https://cdn.jsdelivr.net/npm/three@0.172.0/build/three.module.js",
            "three/addons/": "https://cdn.jsdelivr.net/npm/three@0.172.0/examples/jsm/",
            "d3": "https://cdn.jsdelivr.net/npm/d3@5.16.0/dist/d3.min.js",
        }
    }

settings = Settings()


settings.javascript_location = os.path.join(os.path.dirname(__file__), 'js')

# We default to importing d3 from a CDN
settings.d3_load_string = 'require.config({paths: {d3: "https://cdnjs.cloudflare.com/ajax/libs/d3/5.16.0/d3.min"} });'
# However, if we are working in the pyzx directory itself, we can use the copy of d3
# local to pyzx, which doesn't require an internet connection
# We only do this if we believe we are running in the PyZX directory itself.

try:
    relpath = os.path.relpath(settings.javascript_location, os.getcwd())
    if relpath.count('..') <= 2: # We are *probably* working in the PyZX directory
        settings.javascript_location = os.path.relpath(settings.javascript_location, os.getcwd())
        #settings.d3_load_string = 'require.config({{baseUrl: "{}",paths: {{d3: "d3.v5.min"}} }});'.format(
        #                    settings.javascript_location.replace('\\','/'))
        # TODO: This will fail if Jupyter is started in the parent directory of pyzx, while
        # the notebook is not in the pyzx directory
except ValueError: # relpath raises this Exception when the drive letters don't match
    pass


settings.mode = ''
def get_mode():
    if settings.mode == '':
        try:
            import IPython # type: ignore
            ipython_instance = IPython.get_ipython()
            if ipython_instance is None: raise Exception
            if 'IPKernelApp' in ipython_instance.config:
                settings.mode = "notebook"
                if os.environ.get('SPHINXTARGET') == 'latexpdf':
                    settings.drawing_backend = "matplotlib"
            ipython_instance.config.InlineBackend.figure_format = 'svg'
        except:
            try:
                import browser # type: ignore
                settings.mode = "browser"
            except:
                settings.mode = "shell"

    return settings.mode



def restricted_float(x):
    x = float(x)
    if x < 0.0 or x > 1.0:
        raise ArgumentTypeError("%r not in range [0.0, 1.0]." % (x,))
    return x


def make_into_list(possible_list: Any) -> List[Any]:
    if not isinstance(possible_list, List):
        return [possible_list]
    return possible_list


def maxelements(seq, key=None, reverse=False):
    """
    Return list of position(s) of largest element.

    Adapted from: https://stackoverflow.com/questions/3989016/how-to-find-all-positions-of-the-maximum-value-in-a-list
    """
    indices = []
    if key is None:
        key = lambda x: x
    if reverse:
        compare = lambda x, y: x <= y
    else:
        compare = lambda x, y: x >= y
    if seq:
        best_val = key(seq[0])
        for i, val in enumerate(seq):
            cur_val = key(val)
            if compare(cur_val, best_val):
                if cur_val == best_val:
                    indices.append(i)
                else:
                    best_val = cur_val
                    indices = [i]

    return indices


def is_pauli(phase):
    """
    Check whether phase is Pauli.

    Compatible with zxlive symbols.
    """
    if phase == 0 or phase == 1:
        return True
    return getattr(phase, 'is_pauli', False)

def get_z_box_label(g, v):
    assert g.type(v) == VertexType.Z_BOX
    return g.vdata(v, 'label', 1)

def set_z_box_label(g, v, label):
    assert g.type(v) == VertexType.Z_BOX
    g.set_vdata(v, 'label', label)

# Return position 'perc'%-distance between 2 points:
def ave_pos(a,b,perc=1/2): return (abs(a-b))*(perc) + min(a,b)



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import os
from fractions import Fraction
import traceback

from typing import Optional, List, Tuple, Set, Dict, Any, Union

import pyperclip # type: ignore # Needed for clipboard actions

from .utils import VertexType
from .utils import settings, get_mode, phase_to_s, FloatInt
from .drawing import matrix_to_latex
from .graph import Scalar
from .graph.graph import GraphS
from .rewrite_rules import rules
from . import tikz

from .editor_actions import MATCHES_VERTICES, MATCHES_EDGES, operations, operations_to_js

if get_mode() == 'notebook':
	import ipywidgets as widgets
	from traitlets import Unicode, validate, Bool, Int, Float
	from IPython.display import display, HTML
else:
	# Make some dummy classes to prevent errors with the definition
	# of ZXEditorWidget
	class DOMWidget(object):
		pass
	class Unicode(object): # type: ignore
		def __init__(self,*args,**kwargs):
			pass
		def tag(self, sync=False):
			pass
	class Float(Unicode): # type: ignore
		pass
	class widgets(object): # type: ignore
		register = lambda x: x
		DOMWidget = DOMWidget

__all__ = ['edit', 'help']

HELP_STRING = """To create an editor, call `e = zx.editor.edit(g)` on a graph g. 
This will display the editor, and give you access to 
the underlying Python instance e. Your changes are automatically pushed onto 
the underlying graph instance g (which can also be accessed as e.graph).
Adding the optional argument `zx.editor.edit(g,show_matrix=True)` will
also display the matrix the diagram represents beneath the editor.

Click on edges or vertices to select them. 
Drag a box or hold shift to select multiple vertices or edges.
Press delete or backspace to delete the current selection.
Double-click a vertex to choose its phase.
Ctrl-click (Command-click for mac users) on empty space to add a new vertex. 
The type of the vertex is determined by the box "Vertex type".
Click this box (or press the hotkey 'x') to change the adding type. 
Ctrl-drag between two vertices to add an edge between them. 
The type of edge is determined by the box "Edge type".
Click this box (or press the hotkey 'e') to change the adding type.

When you have a selection, the buttons below the graph light up to denote that 
a rewrite rule can be applied to some of the vertices or edges in this selection.
Press Ctrl+C to copy the selection as tikz code onto the clipboard.
Press Ctrl+V to paste a ZX-diagram on the clipboard (specified as tikz) into the editor.
Press Ctrl+Z to undo an action and Ctrl+Shift+Z to redo an undo'd action.

In order to reflect a change done on g in the Python kernel in the editor itself,
call e.update().
Example:
In [0]: c = zx.Circuit(3)
        c.add_gate("TOF",0,1,2)
        g = c.to_basic_gates().to_graph()
        e = zx.editor_edit(g)
>>> Now the graph g is shown in the output of the cell.
In [1]: zx.spider_simp(g)
        e.update()
>>> Now the view of g in the editor above is updated.
"""

def help():
	print(HELP_STRING)

ERROR_STRING = """This functionality is only supported in a Jupyter Notebook.
If you are running this in a Jupyter notebook, then you probably don't have ipywidgets installed.
Run %%pip install ipywidgets in a cell in your notebook to install the correct package.
"""

def load_js() -> None:
	if get_mode() != 'notebook':
		raise Exception(ERROR_STRING)
	if not widgets.__version__.startswith('7'):
		raise Exception("The editor is only supported with ipywidgets>=7.5,<8. Upgrade or downgrade your install before using it.")
	with open(os.path.join(settings.javascript_location,"zx_editor_widget.js")) as f:
		data1 = f.read()
	with open(os.path.join(settings.javascript_location,"zx_editor_model.js")) as f:
		data2 = f.read()
	#"""<div style="overflow:auto">Loading scripts</div>
	text = """<script type="text/javascript">{0}
				{1}
			</script>
			<script type="text/javascript">
				{2}
			</script>""".format(settings.d3_load_string,data1,data2)
	display(HTML(text))

def s_to_phase(s: str, t:VertexType=VertexType.Z) -> Fraction:
	if not s: 
		if t!= VertexType.H_BOX: return Fraction(0)
		else: return Fraction(1)
	s = s.replace('\u03c0', '')
	if s.find('/') != -1:
		a,b = s.split("/", 2)
		if not a: return Fraction(1,int(b))
		if a == '-': a = '-1'
		return Fraction(int(a),int(b))
	if not s: return Fraction(1)
	return Fraction(int(s))

def graph_to_json(g: GraphS, scale:FloatInt, verts:Optional[List[int]]=None,edges:Optional[List[Tuple[int,int]]]=None) -> str:
	if verts is None:
		verts = list(g.vertices())
	if edges is None:
		edges = list(g.edges())
	nodes = [{'name': int(v),
			  'x': (g.row(v) + 1) * scale,
			  'y': (g.qubit(v) + 2) * scale,
			  't': g.type(v),
			  'phase': phase_to_s(g.phase(v),g.type(v)) }
			 for v in verts]
	links = [{'source': int(g.edge_s(e)),
			  'target': int(g.edge_t(e)),
			  't': g.edge_type(e) } for e in edges]
	scalar = g.scalar.to_json()
	return json.dumps({'nodes': nodes, 'links': links, 'scalar': scalar})


@widgets.register
class ZXEditorWidget(widgets.DOMWidget):
	_view_name = Unicode('ZXEditorView').tag(sync=True)
	_model_name = Unicode('ZXEditorModel').tag(sync=True)
	_view_module = Unicode('zx_editor').tag(sync=True)
	_model_module = Unicode('zx_editor').tag(sync=True)
	_view_module_version = Unicode('0.1.0').tag(sync=True)
	_model_module_version = Unicode('0.1.0').tag(sync=True)
	
	graph_json = Unicode('{"nodes": [], "links": []}').tag(sync=True)
	graph_selected = Unicode('{"nodes": [], "links": []}').tag(sync=True)
	graph_id = Unicode('0').tag(sync=True)
	graph_width = Float(600.0).tag(sync=True)
	graph_height = Float(400.0).tag(sync=True)
	graph_node_size = Float(5.0).tag(sync=True)

	graph_buttons  = Unicode('{empty: false}').tag(sync=True)
	button_clicked = Unicode('').tag(sync=True)
	last_operation = Unicode('').tag(sync=True)
	action 		   = Unicode('').tag(sync=True)
	
	def __init__(
			self, 
			graph: GraphS, 
			show_matrix:bool=False, 
			show_scalar:bool=False,
			*args, **kwargs
			) -> None:
		super().__init__(*args,**kwargs)
		self.observe(self._handle_graph_change, 'graph_json')
		self.observe(self._selection_changed, 'graph_selected')
		self.observe(self._apply_operation, 'button_clicked')
		self.observe(self._perform_action, 'action')
		self.graph = graph
		self.show_matrix = show_matrix
		self.show_scalar = show_scalar
		self.undo_stack: List[Tuple[str,str]] = [('initial',str(self.graph_json))]
		self.undo_position: int = 1
		self.halt_callbacks: bool = False
		self.snapshots: List[GraphS] = []
		self.msg: List[str] = []
		self.output = widgets.Output()
		self.scalar_view = widgets.Label()
		self.matrix_view = widgets.Label()
		self._update_matrix()
	
	def update(self) -> None:
		self.graph_json = graph_to_json(self.graph, self.graph.scale) # type: ignore

	def _update_matrix(self):
		if self.show_scalar:
			s = self.graph.scalar.to_latex()
			if s == '': s = '1'
			self.scalar_view.value = "Scalar: " + s
		if not self.show_matrix: return
		try:
			self.graph.auto_detect_io()
		except TypeError:
			self.matrix_view.value = "Couldn't parse inputs or outputs"
			return
		if self.graph.num_inputs() > 4 or self.graph.num_outputs() > 4:
			self.matrix_view.value = "Matrix too large to show"
			return
		try:
			m = self.graph.to_matrix()
		except ValueError:
			return
		self.matrix_view.value = matrix_to_latex(m)

	def _parse_selection(self) -> Tuple[Set[int],Set[Tuple[int,int]]]:
		"""Helper function for `_selection_changed` and `_apply_operation`."""
		selection = json.loads(self.graph_selected)
		g = self.graph
		vertex_set = set([n["name"] for n in selection["nodes"]])
		edge_set = set([g.edge(e["source"],e["target"]) for e in selection["links"]])
		edge_set.update([g.edge(v,w) for v in vertex_set for w in vertex_set if g.connected(v,w)])
		return vertex_set, edge_set

	def _selection_changed(self, change):
		"""Is called when the selection in the editor changes.
		Updates the action buttons so that the correct ones are active."""
		try:
			vertex_set, edge_set = self._parse_selection()
			g = self.graph
			js = json.loads(self.graph_buttons)
			for op_id, data in operations.items():
				if data["type"] == MATCHES_EDGES:
					matches = data["matcher"](g, lambda e: e in edge_set)
				else: matches = data["matcher"](g, lambda v: v in vertex_set)
				js[op_id]["active"] = (len(matches) != 0)
			self.graph_buttons = json.dumps(js)
		except Exception as e:
			with self.output: print(traceback.format_exc())

	def _apply_operation(self, change):
		"""called when one of the action buttons is clicked. 
		Performs the action on the selection."""
		try:
			op = change['new']
			if not op: return
			vertex_set, edge_set = self._parse_selection()
			g = self.graph
			data = operations[op]
			if data["type"] == MATCHES_EDGES:
				matches = data["matcher"](g, lambda e: e in edge_set)
			else: matches = data["matcher"](g, lambda v: v in vertex_set)
			# Apply the rule
			etab, rem_verts, rem_edges, check_isolated_vertices = data["rule"](g, matches)
			g.remove_edges(rem_edges)
			g.remove_vertices(rem_verts)
			g.add_edge_table(etab)
			
			#if check_isolated_vertices: g.remove_isolated_vertices()
			# Remove stuff from the selection
			selection = json.loads(self.graph_selected)
			selection["nodes"] = [v for v in selection["nodes"] if v["name"] not in rem_verts]
			selection["links"] = [e for e in selection["links"] if (
										(e["source"], e["target"]) not in rem_edges 
										and e["source"] not in rem_verts
										and e["target"] not in rem_verts)]
			self.graph_selected = json.dumps(selection)
			self.button_clicked = ''
			self.update()
			self._selection_changed(None)
		except Exception as e:
			with self.output: print(traceback.format_exc())

	def _perform_action(self, change):
		try:
			action = change['new']
			if action == '': return
			elif action == 'undo': self.undo()
			elif action == 'redo': self.redo()
			elif action == 'snapshot': self.make_snapshot()
			elif action == 'tikzit': self.open_tikzit()
			elif action == 'paste': self.paste()
			elif action == 'copy': self.copy_to_clipboard()
			else: raise ValueError("Unknown action '{}'".format(action))
			self.action = ''
		except Exception as e:
			with self.output: print(traceback.format_exc())


	def _undo_stack_add(self, description: str, js: str) -> None:
		self.undo_stack = self.undo_stack[:len(self.undo_stack)-self.undo_position+1]
		self.undo_position = 1
		self.undo_stack.append((description,js))
		self.msg.append("Adding to undo stack: " + description)

	def undo(self) -> None:
		if self.undo_position == len(self.undo_stack): return
		self.undo_position += 1
		description, js = self.undo_stack[len(self.undo_stack)-self.undo_position]
		self.msg.append("Undo {}: {:d}-{:d}".format(description,len(self.undo_stack),self.undo_position))
		self.halt_callbacks = True
		self.graph_selected = '{"nodes": [], "links": []}'
		self.graph_from_json(json.loads(js))
		self.update()
		self.halt_callbacks = False

	def redo(self) -> None:
		if self.undo_position == 1: return
		self.undo_position -= 1
		description, js = self.undo_stack[len(self.undo_stack)-self.undo_position]
		self.msg.append("Redo {}: {:d}-{:d}".format(description,len(self.undo_stack),self.undo_position))
		self.halt_callbacks = True
		self.graph_selected = '{"nodes": [], "links": []}'
		self.graph_from_json(json.loads(js))
		self.update()
		self.halt_callbacks = False

	def paste(self) -> None:
		v = pyperclip.paste()
		try:
			g = tikz.tikz_to_graph(v,warn_overlap=False,ignore_nonzx=True)
		except ValueError as e:
			self.msg.append("Tried loading tikz from clipboard, but failed:")
			self.msg.append(str(e))
			return
		self.msg.append("Copied graph from clipboard")
		minrow = min([g.row(v) for v in g.vertices()], default=0)
		minqub = min([g.qubit(v) for v in g.vertices()], default=0)
		if minrow > 0: minrow = 0
		if minqub > 0: minqub = 0
		minrow = minrow - 0.3
		minqub = minqub - 0.3
		g = g.translate(-minrow,-minqub)
		self.halt_callbacks = True
		verts, edges = self.graph.merge(g)
		# self.graph_selected = '{"nodes": [], "links": []}'
		# js = graph_to_json(g, self.graph.scale) # type: ignore
		# self.graph_from_json(json.loads(js))
		self._undo_stack_add("paste",graph_to_json(self.graph,self.graph.scale)) # type: ignore
		self.update()
		self.graph_selected = graph_to_json(self.graph,self.graph.scale,verts,[]) # type: ignore
		self.halt_callbacks = False
		# self._selection_changed(None)

	def copy_to_clipboard(self) -> None:
		selection = json.loads(self.graph_selected)
		verts = [n["name"] for n in selection["nodes"]]
		g = self.graph.subgraph_from_vertices(verts)
		tz = g.to_tikz()
		self.msg.append("Copying graph to clipboard")
		pyperclip.copy(tz)
		

	def make_snapshot(self) -> None:
		self.snapshots.append(self.graph.copy()) # type: ignore

	def open_tikzit(self) -> None:
		seq = self.snapshots + [self.graph]
		tz = tikz.to_tikz_sequence(seq) # type: ignore
		try:
			tikz.tikzit(tz)
		except Exception as e:
			with self.output: print(e)

	def graph_from_json(self, js: Dict[str,Any]) -> None:
		try:
			scale = self.graph.scale # type: ignore
			marked: Union[Set[int],Set[Tuple[int,int]]] = self.graph.vertex_set()
			for n in js["nodes"]:
				v = n["name"]
				r = float(n["x"])/scale -1
				q = float(n["y"])/scale -2
				t = int(n["t"])
				phase = s_to_phase(n["phase"], t) # type: ignore
				if v not in marked:
					self.graph.add_vertex_indexed(v)
				else: 
					marked.remove(v)
				self.graph.set_position(v, q, r)
				self.graph.set_phase(v, phase)
				self.graph.set_type(v, t)
			self.graph.remove_vertices(marked)
			marked = self.graph.edge_set()
			for e in js["links"]:
				s = int(e["source"])
				t = int(e["target"])
				et = int(e["t"])
				if self.graph.connected(s,t):
					f = self.graph.edge(s,t)
					marked.remove(f)
					self.graph.set_edge_type(f, et)
				else:
					self.graph.add_edge((s,t),et) # type: ignore
			self.graph.remove_edges(marked)
			if 'scalar' in js:
				self.graph.scalar = Scalar.from_json(js['scalar'])
			self._update_matrix()
		except Exception as e:
			with self.output: print(traceback.format_exc())
	
	def _handle_graph_change(self, change):
		"""Called whenever the graph in the editor is modified."""
		if self.halt_callbacks: return
		self.msg.append("Handling graph change")
		try:
			js = json.loads(change['new'])
			js['scalar'] = self.graph.scalar.to_json()
			self.graph_from_json(js)
			self._undo_stack_add(self.last_operation, json.dumps(js))
		except Exception as e:
			with self.output: print(traceback.format_exc())
		

	def to_graph(self, zh:bool=True) -> GraphS:
		return self.graph

			

_d3_editor_id = 0

def edit(
		g: GraphS, 
		scale:Optional[FloatInt]=None, 
		show_matrix:bool=False,
		show_scalar:bool=False,
		show_errors:bool=True) -> ZXEditorWidget:
	"""Start an instance of an ZX-diagram editor on a given graph ``g``.
	Only usable in a Jupyter Notebook. 
	When this function is called it displays a Jupyter Widget that allows
	you to modify a pyzx Graph instance in the Notebook itself.
	This function returns an instance of the editor widget, so it should be called like::

		e = zx.editor.edit(g)

	Usage:
		Ctrl-click on empty space to add vertices.
		Ctrl-drag between vertices to add edges.
		Use the "Vertex type" and "Edge type" buttons to toggle which type of
		vertex or edge to add.
		Drag with left-mouse to make a selection.
		Left-drag on a vertex to move it.
		Delete or backspace removes the selection.
		Ctrl-Z and Ctrl-Shift-Z undoes and redoes the last action.
		With a part of the graph selected, click one of the action buttons
		beneath the graph to perform a ZX-calculus rewrite.
		Click "Save snapshot" to store the current graph into ``e.snapshots``.
		Click "Load in tikzit" to open all snapshots in Tikzit.
		Point ``zx.settings.tikzit_location`` to a Tikzit executable to use this function.

	Args:
		g: The Graph instance to edit
		scale: What size the vertices should have (ideally somewhere between 20 and 50)
		show_matrix: When True, displays the linear map the Graph implements beneath the editor
		show_scalar: When True, displays ``g.scalar`` beneath the editor.
		show_errors: When True, prints Exceptions beneath the editor
	
	"""
	load_js()
	global _d3_editor_id
	_d3_editor_id += 1
	seq = _d3_editor_id

	if scale is None:
		scale = 800 / (g.depth() + 2)
		if scale > 50: scale = 50
		if scale < 20: scale = 20
	
	g.scale = scale # type: ignore
	
	node_size = 0.2 * scale
	if node_size < 2: node_size = 2

	w = max([(g.depth() + 2) * scale, 400])
	h = max([(g.qubit_count() + 3) * scale + 30, 200])
	
	js = graph_to_json(g, scale)


	w = ZXEditorWidget(
					g, show_matrix,show_scalar,
					graph_json = js, graph_id = str(seq), 
					graph_width=w, graph_height=h, 
					graph_node_size=node_size,
					graph_buttons = operations_to_js()
					)
	display(w)
	if show_scalar:
		display(w.scalar_view)
	if show_matrix:
		display(w.matrix_view)
	if show_errors: display(w.output)
	return w



# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import os
import sys
from types import ModuleType
from typing import Optional

if __name__ == '__main__':
    sys.path.append('..')
    sys.path.append('.')
from pyzx_param.simplify import full_reduce
from pyzx_param.extract import extract_circuit
from pyzx_param.circuit import Circuit
from fractions import Fraction

np: Optional[ModuleType]
try:
    import numpy as np
    from pyzx_param.tensor import compare_tensors
    import math
except ImportError:
    np = None

try:
    from qiskit import quantum_info, transpile
    from qiskit.circuit import QuantumCircuit
    from qiskit.qasm2 import dumps as dumps2
    from qiskit.qasm3 import loads as loads3, dumps as dumps3
except ImportError:
    QuantumCircuit = None


@unittest.skipUnless(np, "numpy needs to be installed for this to run")
class TestQASM(unittest.TestCase):

    def test_parser_state_reset(self):
        from pyzx_param.circuit.qasmparser import QASMParser
        s = """
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[1];
        h q[0];
        """
        p = QASMParser()
        c1 = p.parse(s)
        c2 = p.parse(s)
        self.assertEqual(c2.qubits, 1)
        self.assertEqual(len(c2.gates), 1)
        self.assertTrue(c1.verify_equality(c2))

    def test_parse_qasm3(self):
        qasm3 = Circuit.from_qasm("""
        OPENQASM 3;
        include "stdgates.inc";
        qubit[3] q;
        cx q[0], q[1];
        s q[2];
        cx q[2], q[1];
        """)
        c = Circuit(3)
        c.add_gate("CNOT", 0, 1)
        c.add_gate("S", 2)
        c.add_gate("CNOT", 2, 1)
        self.assertEqual(c.qubits, qasm3.qubits)
        self.assertListEqual(c.gates, qasm3.gates)


    def test_parse_qasm3_long_creg(self):
        qasm3 = Circuit.from_qasm("""
        OPENQASM 3;
        include "stdgates.inc";
        qubit[3] q1;
        cx q1[0], q1[1];
        s q1[2];
        cx q1[2], q1[1];
        """)
        c = Circuit(3)
        c.add_gate("CNOT", 0, 1)
        c.add_gate("S", 2)
        c.add_gate("CNOT", 2, 1)
        self.assertEqual(c.qubits, qasm3.qubits)
        self.assertListEqual(c.gates, qasm3.gates)

    def test_load_qasm_from_file(self):
        c = Circuit(1)
        c.add_gate("YPhase", 0, Fraction(1, 4))
        c1 = Circuit.from_qasm_file(os.path.join(os.path.dirname(__file__), "ry.qasm"))
        self.assertEqual(c1.qubits, c.qubits)
        self.assertListEqual(c1.gates,c.gates)

    def test_broadcasting(self):
        """Test that broadcasting is handled correctly.

        If any arguments of a gate are quantum registers instead of qubits, it is a shorthand for broadcasting
        over the qubits of the register."""
        c1 = Circuit.from_qasm("""
        OPENQASM 3;
        include "stdgates.inc";
        qubit[1] q0;
        qubit[3] q1;
        cx q0[0], q1;
        """)
        c2 = Circuit(4)
        c2.add_gate("CNOT", 0, 1)
        c2.add_gate("CNOT", 0, 2)
        c2.add_gate("CNOT", 0, 3)
        self.assertEqual(c1.qubits, c2.qubits)
        self.assertListEqual(c1.gates, c2.gates)

    def test_catch_broadcasting_error(self):
        """Test that all registers are of the same length when broadcasting."""
        with self.assertRaises(TypeError) as context:
            Circuit.from_qasm("""
            OPENQASM 3;
            include "stdgates.inc";
            qubit[1] q0;
            qubit[2] q1;
            qubit[3] q2;
            ccx q0[0], q1, q2;
            """)
        self.assertTrue("Register sizes do not match" in str(context.exception))

    def test_p_same_as_rz(self):
        """Test that the `p` gate is identical to the `rz` gate.

        In the OpenQASM 3 spec, they differ by a global phase. However,
        they are treated in pyzx identically, since the global phase doesn't
        matter.
        """
        p_qasm = Circuit.from_qasm("""
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[1];
        p(pi/2) q[0];
        """)
        rz_qasm = Circuit.from_qasm("""
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[1];
        rz(pi/2) q[0];
        """)
        self.assertEqual(p_qasm.qubits, rz_qasm.qubits)
        self.assertListEqual(p_qasm.gates, rz_qasm.gates)

    def test_cp_differs_from_crz(self):
        """Test that the `cp` gate and the `crz` gates are different.

        The difference in phase between `p` and `rz` matters when the gates are
        controlled.
        """
        crz_qasm = Circuit.from_qasm("""
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[2];
        crz(pi/2) q[0],q[1];
        """)
        cp_qasm = Circuit.from_qasm("""
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[2];
        cp(pi/2) q[0],q[1];
        """)
        crz_t = crz_qasm.to_matrix()
        cp_t = cp_qasm.to_matrix()
        self.assertFalse(compare_tensors(crz_t, cp_t, False))

    def test_rzz(self):
        """Regression test for issue #158.

        The qiskit transpiler may rewrite `crz` into `rz` and `rzz` in some
        situations. The following circuits should be equivalent."""
        before = Circuit.from_qasm("""
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[2];
        crz(pi/2) q[0],q[1];
        """)

        after = Circuit.from_qasm("""
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[2];
        rz(pi/4) q[1];
        rzz(-pi/4) q[0],q[1];
        """)
        self.assertTrue(compare_tensors(
            before.to_matrix(), after.to_matrix(), True))

    def test_custom_gates(self):
        from pyzx_param.circuit.qasmparser import QASMParser
        s1 = """
        OPENQASM 2.0;
        include "qelib1.inc";
        gate majority a,b,c {
            cx c,b;
            cx c,a;
            ccx a,b,c;
        }
        qreg q[3];
        majority q[0],q[1],q[2];
        """
        s2 = """
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[3];
        cx q[2],q[1];
        cx q[2],q[0];
        ccx q[0],q[1],q[2];
        """
        p = QASMParser()
        c1 = p.parse(s1)
        c2 = p.parse(s2)
        self.assertEqual(c1.qubits, c2.qubits)
        self.assertListEqual(c1.gates, c2.gates)

    @unittest.skipUnless(QuantumCircuit, "qiskit needs to be installed for this test")
    def test_qasm_qiskit_semantics(self):
        """Verify/document qasm gate semantics when imported into pyzx.

        pyzx's implementation of qasm gates differs from qiskit in global
        phases which are not significant, but which has sometimes led to user
        confusion. This unit test documents those differences.
        """

        def compare_gate_matrix_with_qiskit(gates, num_qubits: int, num_angles: int, qasm_versions=None):
            for gate in gates:
                for qasm_version in qasm_versions if qasm_versions else [2, 3]:
                    header = "OPENQASM 2.0;\ninclude \"qelib1.inc\";\n" if qasm_version == 2 \
                        else "OPENQASM 3;\ninclude \"stdgates.inc\";\n"
                    params = "" if num_angles == 0 else "({})".format(
                        ",".join(["pi/2"] * num_angles))
                    setup = header + f"qreg q[{num_qubits}];\n{gate}{params} "
                    qasm = setup + \
                        ", ".join(
                            [f"q[{i}]" for i in range(num_qubits)]) + ";\n"
                    pyzx_circuit = Circuit.from_qasm(qasm)
                    pyzx_matrix = pyzx_circuit.to_matrix()

                    for g in pyzx_circuit.gates:
                        for b in g.to_basic_gates():
                            self.assertListEqual(b.to_basic_gates(), [b],
                                                 f"\n{gate}.to_basic_gates() contains non-basic gate")

                    # qiskit uses little-endian ordering
                    qiskit_qasm = setup + \
                        ", ".join([f"q[{i}]" for i in reversed(
                            range(num_qubits))]) + ";\n"
                    qc = QuantumCircuit.from_qasm_str(qiskit_qasm) if qasm_version == 2 else loads3(qiskit_qasm)
                    qiskit_matrix = quantum_info.Operator(qc).data

                    # Check that pyzx and qiskit produce the same tensor from the same qasm, modulo qubit endianness.
                    self.assertTrue(compare_tensors(pyzx_matrix, qiskit_matrix, False),
                                    f"Gate: {gate}\nqasm:\n{qasm}\npyzx_matrix:\n{pyzx_matrix}\nqiskit_matrix:\n{qiskit_matrix}")

                    # Check internal round-trip (pyzx to qasm to pyzx) results in the same circuit.
                    qasm_from_pyzx = pyzx_circuit.to_qasm(qasm_version)
                    pyzx_round_trip = Circuit.from_qasm(qasm_from_pyzx)
                    self.assertEqual(pyzx_circuit.qubits, pyzx_round_trip.qubits)
                    self.assertListEqual(pyzx_circuit.gates, pyzx_round_trip.gates)

                    # Check external round-trip (pyzx to qasm to qiskit to qasm to pyzx) results in the same circuit.
                    # Note that the endianness is reversed when going out and again when coming back in, so the overall
                    # result is no change.
                    qiskit_from_qasm = (QuantumCircuit.from_qasm_str(qasm_from_pyzx) if qasm_version == 2
                                        else loads3(qasm_from_pyzx))
                    pyzx_from_qiskit = Circuit.from_qasm(dumps2(qiskit_from_qasm) if qasm_version == 2
                                                         else dumps3(qiskit_from_qasm))
                    self.assertEqual(pyzx_circuit.qubits, pyzx_from_qiskit.qubits)
                    self.assertListEqual(pyzx_circuit.gates, pyzx_from_qiskit.gates)

        # Test standard gates common to both qelib1.inc (OpenQASM 2) and stdgates.inc (OpenQASM 3).
        compare_gate_matrix_with_qiskit(
            ['x', 'y', 'z', 'h', 's', 'sdg', 't', 'tdg', 'sx'], 1, 0)
        compare_gate_matrix_with_qiskit(['u1', 'p', 'rx', 'ry', 'rz'], 1, 1)
        compare_gate_matrix_with_qiskit(['u2'], 1, 2)
        compare_gate_matrix_with_qiskit(['u3'], 1, 3)
        compare_gate_matrix_with_qiskit(
            ['cx', 'CX', 'cy', 'cz', 'ch', 'swap'], 2, 0)
        compare_gate_matrix_with_qiskit(['crx', 'cry', 'crz', 'cp'], 2, 1)
        # 'ccz' not tested because not a standard qasm gate
        compare_gate_matrix_with_qiskit(['ccx', 'cswap'], 3, 0)
        compare_gate_matrix_with_qiskit(['cu'], 2, 4)

        # Test standard gates added to stdgates.inc.
        compare_gate_matrix_with_qiskit(['cphase'], 2, 1, [3])

        # Test standard gates removed from stdgates.inc.
        compare_gate_matrix_with_qiskit(['sxdg'], 1, 0, [2])
        compare_gate_matrix_with_qiskit(['csx'], 2, 0, [2])
        compare_gate_matrix_with_qiskit(['cu1', 'rxx', 'rzz'], 2, 1, [2])
        compare_gate_matrix_with_qiskit(['cu3'], 2, 3, [2])
        compare_gate_matrix_with_qiskit(['u'], 1, 3, [2])

        # Test native OpenQASM 3 gate.
        compare_gate_matrix_with_qiskit(['U'], 1, 3, [3])

    @unittest.skipUnless(QuantumCircuit, "qiskit needs to be installed for this test")
    def test_qiskit_transpile_pyzx_optimization_round_trip(self):
        """Regression test for issue #102.

        Transpile a circuit in qiskit, simplify it using pyzx, and verify that
        it produces the same unitary in qiskit.
        """
        qc = QuantumCircuit(4)
        qc.ccx(2, 1, 0)
        qc.ccz(0, 1, 2)
        qc.h(1)
        qc.ccx(1, 2, 3)
        qc.t(1)
        qc.ccz(0, 1, 2)
        qc.h(1)
        qc.t(0)
        qc.ccz(2, 1, 0)
        qc.s(1)
        qc.ccx(2, 1, 0)
        qc.crz(0.2*np.pi, 0, 1)
        qc.rz(0.8*np.pi, 1)
        qc.cry(0.4*np.pi, 2, 1)
        qc.crx(0.02*np.pi, 2, 0)

        qc1 = transpile(qc)
        t1 = quantum_info.Operator(qc1).data

        # Test round-trip for OpenQASM 2.
        c2 = Circuit.from_qasm(dumps2(qc1))
        g2 = c2.to_graph()
        full_reduce(g2)
        qasm2 = extract_circuit(g2).to_basic_gates().to_qasm(2)

        qc2 = QuantumCircuit().from_qasm_str(qasm2)
        t2 = quantum_info.Operator(qc2).data

        self.assertTrue(compare_tensors(t1, t2))

        # Test round-trip for OpenQASM 3.
        c3 = Circuit.from_qasm(dumps3(qc1))
        g3 = c3.to_graph()
        full_reduce(g3)
        qasm3 = extract_circuit(g3).to_basic_gates().to_qasm(3)

        qc3 = loads3(qasm3)
        t3 = quantum_info.Operator(qc3).data

        self.assertTrue(compare_tensors(t1, t3))


if __name__ == '__main__':
    unittest.main()



# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import random
import sys
import os
from types import ModuleType
from typing import Optional

if __name__ == '__main__':
    sys.path.append('..')
    sys.path.append('.')
mydir = os.path.dirname(__file__)
from pyzx_param.generate import cliffordT, cliffords
from pyzx_param.simplify import clifford_simp
from pyzx_param.extract import extract_circuit
from pyzx_param.circuit import Circuit

np: Optional[ModuleType]
try:
    import numpy as np
    from pyzx_param.tensor import tensorfy, compare_tensors
    import math
except ImportError:
    np = None

SEED = 1337


@unittest.skipUnless(np, "numpy needs to be installed for this to run")
class TestCircuit(unittest.TestCase):

    def setUp(self):
        c = Circuit(3)
        c.add_gate("CNOT",0,1)
        c.add_gate("S",2)
        c.add_gate("CNOT",2,1)
        self.c = c

    def test_to_graph_and_back(self):
        g = self.c.to_graph()
        c2 = Circuit.from_graph(g)
        self.assertEqual(self.c.qubits, c2.qubits)
        self.assertListEqual(self.c.gates,c2.gates)

    def test_to_qasm_and_back(self):
        s = self.c.to_qasm()
        c2 = Circuit.from_qasm(s)
        self.assertEqual(self.c.qubits, c2.qubits)
        self.assertListEqual(self.c.gates,c2.gates)

    def test_to_qc_and_back(self):
        s = self.c.to_qc()
        c2 = Circuit.from_qc(s)
        self.assertEqual(self.c.qubits, c2.qubits)
        self.assertListEqual(self.c.gates,c2.gates)

    def test_to_quipper_and_back(self):
        s = self.c.to_quipper()
        c2 = Circuit.from_quipper(s)
        self.assertEqual(self.c.qubits, c2.qubits)
        self.assertListEqual(self.c.gates,c2.gates)

    def test_load_quipper_from_file(self):
        c1 = Circuit.from_quipper_file(os.path.join(mydir,"test_circuit.circuit"))
        c2 = Circuit.from_quipper_file(os.path.join(mydir,"test_circuit_nocontrol_noqubits.circuit"))
        self.assertEqual(c1.qubits, c2.qubits)
        self.assertListEqual(c2.gates,c2.gates)

    def test_cliffordT_preserves_graph_semantics(self):
        for i in range(1,5):
            with self.subTest(i):
                random.seed(SEED)
                g = cliffordT(i,20,0.2)
                c = Circuit.from_graph(g)
                g2 = c.to_graph()
                t = tensorfy(g,False)
                t2 = tensorfy(g2,False)
                self.assertTrue(compare_tensors(t,t2, False))

    def test_cliffordT_raises_exception_with_pcnot(self):
        self.assertRaises(ValueError, cliffordT, 1, 1, p_cnot=0.1)

    def test_cliffords_preserves_graph_semantics(self):
        random.seed(SEED)
        g = cliffords(5,30)
        c = Circuit.from_graph(g)
        g2 = c.to_graph()
        t = tensorfy(g,False)
        t2 = tensorfy(g2,False)
        self.assertTrue(compare_tensors(t,t2,False))

    def test_circuit_extract_preserves_semantics(self):
        random.seed(SEED)
        g = cliffordT(5, 70, 0.15)
        t = g.to_tensor(False)
        clifford_simp(g, quiet=True)
        c = extract_circuit(g)
        t2 = c.to_tensor(False)
        self.assertTrue(compare_tensors(t,t2,False))

    def test_two_qubit_gate_semantics(self):
        c = Circuit(2)
        c.add_gate("CNOT",0,1)
        cnot_matrix = np.array([[1,0,0,0],[0,1,0,0],[0,0,0,1],[0,0,1,0]])
        self.assertTrue(compare_tensors(c.to_matrix(),cnot_matrix))
        c = Circuit(2)
        c.add_gate("CZ",0,1)
        cz_matrix = np.array([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,-1]])
        self.assertTrue(compare_tensors(c.to_matrix(),cz_matrix))

    def test_measurement_gate(self):
        c = Circuit(2) 
        c1 = Circuit(2)
        c.add_gate("Measurement", 0)
        c1.add_gate("Measurement", 0, None)
        
        self.assertTrue(c.gates[0] == c1.gates[0])
        g = c.to_graph()
        g1 = c1.to_graph()
        self.assertTrue(len(g.vertices()) == len(g1.vertices()))


    def test_verify_equality_permutation_option(self):
        c1 = Circuit(2)
        c2 = Circuit(2)
        c2.add_gate("SWAP",0,1)
        self.assertTrue(c1.verify_equality(c2,up_to_swaps=True))
        self.assertFalse(c1.verify_equality(c2,up_to_swaps=False))

if __name__ == '__main__':
    unittest.main()



# PyZX - Python library for quantum circuit rewriting
#        and optimisation using the ZX-calculus
# Copyright (C) 2021 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
This module contains objective functions to guide local search over ZX-diagrams. The wgc method defines a measure of circuit complexity -- a weighted gate count where 2-qubit counts incur a higher cost. The g_wgc takes a ZX-diagram as input and optionally applies various optimizations before measuring the complexity of the circuit obtained via extraction.
"""

import sys
if __name__ == '__main__':
    sys.path.append('..')
from pyzx_param.extract import extract_circuit
from pyzx_param.simplify import full_reduce
from pyzx_param.optimize import basic_optimization

# Weighted gate count
def wgc(c, two_qb_weight=10):
    """A measure of the complexity of a given circuit.

    :param c: Circuit to evaluate.
    :param two_qb_weight: Weight factor for 2-qubit gates, default is 10.
    :return: Weighted gate count of the circuit.
    """

    c_tmp = c.to_basic_gates()
    total = len(c_tmp.gates)
    n2 = c_tmp.twoqubitcount()
    single_qubit_count = total - n2
    return two_qb_weight * n2 + single_qubit_count

# Weighted gate count of a ZX-diagram
def g_wgc(g, two_qb_weight=10, g_simplify=False, c_simplify=True):
    """A measure of the complexity of the circuit obtained from a a ZX-diagram

    :param g: ZX-diagram to evaluate.
    :param two_qb_weight: Weight factor for 2-qubit gates, default is 10.
    :param g_simplify: If True, applies full reduction to graph before returning weighted count.
    :param c_simplify: If True, applies basic optimization to circuit before returning weighted count.
    :return: Weighted gate count of the extracted circuit."""

    g_tmp = g.copy()
    if g_simplify:
        full_reduce(g_tmp)

    c = extract_circuit(g_tmp).to_basic_gates()

    if c_simplify:
        c = basic_optimization(c)

    return wgc(c, two_qb_weight=two_qb_weight)



# PyZX - Python library for quantum circuit rewriting
#        and optimisation using the ZX-calculus
# Copyright (C) 2019 - Aleks Kissinger, John van de Wetering,
#                      and Arianne Meijer-van de Griend

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import argparse
import os
from ..utils import make_into_list
from ..routing.parity_maps import CNOT_tracker
from ..generate import build_random_parity_map

description = "Generates random CNOT circuits and stores them as QASM files."

if __name__ == "__main__":
    print("Please call this as python -m pyzx cnots ...")
    exit()

parser = argparse.ArgumentParser(description=description)
parser.add_argument(
    "folder", help="The QASM file or folder with QASM files to be routed."
)
parser.add_argument(
    "-q",
    "--n_qubits",
    nargs="+",
    default=9,
    type=int,
    help="The number of qubits participating in the circuit.",
)
parser.add_argument(
    "-m",
    "--n_maps",
    default=1,
    type=int,
    help="The number of circuits to be generated.",
)
parser.add_argument(
    "-d",
    "--n_cnots",
    nargs="+",
    default=None,
    type=int,
    help="The number of CNOTs in the generated circuit.",
)


def main(args):
    args = parser.parse_args()
    if args.n_cnots is None:
        parser.error(
            message="Please specify the number of CNOT gates to be generated with the -d flag."
        )
    folder = args.folder
    os.makedirs(folder, exist_ok=True)

    n_qubits = make_into_list(args.n_qubits)
    n_maps = args.n_maps
    n_cnots = make_into_list(args.n_cnots)

    for q in n_qubits:
        for n in n_cnots:
            dest_folder = os.path.join(folder, str(q) + "qubits", str(n))
            os.makedirs(dest_folder, exist_ok=True)
            for i in range(n_maps):
                filename = "Original" + str(i) + ".qasm"
                dest_file = os.path.join(dest_folder, filename)
                circuit = CNOT_tracker(q)
                build_random_parity_map(q, n, circuit)
                with open(dest_file, "w") as f:
                    f.write(circuit.to_qasm())



# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import random
import sys
import numpy as np
from fractions import Fraction

if __name__ == '__main__':
    sys.path.append('..')
    sys.path.append('.')

import pyzx_param as zx
from pyzx_param.rank_width import tensorfy_rw, conv_uv, conv_vw, conv_naive


class TestRankWidth(unittest.TestCase):
    @staticmethod
    def generate_graph(n, m):
        g = zx.Graph()
        for _ in range(n):
            g.add_vertex(zx.VertexType.Z, phase=Fraction(random.randint(0, 7), 4))
        while g.num_edges() < m:
            [u, v] = random.sample(range(n), k=2)
            if (u, v) not in g.edge_set() and (v, u) not in g.edge_set():
                g.add_edge((u, v), zx.EdgeType.HADAMARD)
        return g

    def check_amplitude(self, res, corr, info=''):
        if not np.allclose(res, corr):
            print(f'Amplitude mismatch: expected {corr}, found {res}')
            print(info)
            self.assertTrue(False)

    def check_graph_tensorfy(self, g, strategy='rw-auto'):
        res = tensorfy_rw(g, strategy=strategy)
        corr = zx.tensorfy(g, strategy='naive')
        self.check_amplitude(res, corr, str(g.edge_set()))

    def check_circuit_tensorfy(self, circ, state, effect, strategy='rw-auto'):
        g = circ.to_graph()
        g.apply_state(state)
        g.apply_effect(effect)
        self.check_graph_tensorfy(g, strategy=strategy)

    def test_convolution(self):
        r_u, r_v, r_w = 2, 3, 4
        Psi_v = np.random.random((2, 2 ** r_v)).astype(np.complex128)
        Psi_w = np.random.random((2, 2 ** r_w)).astype(np.complex128)
        E_vw = np.random.randint(2, size=(r_v, r_w)).astype(np.int8)
        E_vu = np.random.randint(2, size=(r_v, r_u)).astype(np.int8)
        E_wu = np.random.randint(2, size=(r_w, r_u)).astype(np.int8)
        corr = conv_naive(Psi_v, Psi_w, E_vw, E_vu, E_wu)
        res_vw = conv_vw(Psi_v, Psi_w, E_vw, E_vu, E_wu)
        res_uv = conv_uv(Psi_v, Psi_w, E_vw, E_vu, E_wu)
        self.check_amplitude(res_vw, corr, 'conv_vw')
        self.check_amplitude(res_uv, corr, 'conv_uv')

    def test_tensorfy_rw_one_edge(self):
        g = zx.Graph()
        g.add_vertex(zx.VertexType.Z, phase=0)
        g.add_vertex(zx.VertexType.Z, phase=1)
        g.add_edge((0, 1), zx.EdgeType.HADAMARD)
        self.check_graph_tensorfy(g)

    def test_tensorfy_rw_square(self):
        g = zx.Graph()
        for i in range(4):
            g.add_vertex(zx.VertexType.Z)
        for i in range(4):
            g.add_edge((i, (i + 1) % 4), zx.EdgeType.HADAMARD)
        self.check_graph_tensorfy(g)

    def test_tensorfy_rw_NOT(self):
        circ = zx.Circuit(1)
        circ.add_gate('NOT', 0)
        self.check_circuit_tensorfy(circ, '/', '/')

    def test_tensorfy_rw_T(self):
        circ = zx.Circuit(1)
        circ.add_gate('T', 0)
        self.check_circuit_tensorfy(circ, '/', '/')

    def test_tensorfy_rw_random_graph(self):
        for _ in range(10):
            g = self.generate_graph(10, 40)
            self.check_graph_tensorfy(g)

    def test_tensorfy_rw_random_circuit(self):
        n_qubits, n_gates = 10, 200
        basis_states = ['0', '1', '+', '-', '/']
        for _ in range(10):
            circ = zx.generate.CNOT_HAD_PHASE_circuit(qubits=n_qubits, depth=n_gates)
            state = ''.join(random.choice(basis_states) for _ in range(n_qubits))
            effect = ''.join(random.choice(basis_states) for _ in range(n_qubits))
            self.check_circuit_tensorfy(circ, state, effect)

    def test_tensorfy_rw_small_circuit(self):
        self.check_circuit_tensorfy(zx.generate.CNOT_HAD_PHASE_circuit(qubits=1, depth=0),
                                    '0', '0')
        self.check_circuit_tensorfy(zx.generate.CNOT_HAD_PHASE_circuit(qubits=1, depth=1, p_t=0.8),
                                    '+', '+')
        self.check_circuit_tensorfy(zx.generate.CNOT_HAD_PHASE_circuit(qubits=2, depth=0),
                                    '1-', '-1')
        self.check_circuit_tensorfy(zx.generate.CNOT_HAD_PHASE_circuit(qubits=2, depth=1),
                                    '--', '11')


if __name__ == '__main__':
    unittest.main()


# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the ZX-diagram rewrite class of PyZX.
It is part of a major reworking of how the rule and simplify files 
work to perform the rewrite rules on diagrams."""

from typing import List, Callable, Optional, Union, Generic, Set, Tuple, Dict, Iterator, cast
from .graph.base import BaseGraph, VT, ET

class Rewrite(Generic[VT, ET]):

    def __init__(self) -> None:
        pass

    def simp(self, graph) -> bool:
        raise Exception("This rewrite rule cannot terminate when run automatically. Try using apply() instead to manually target vertices.")

    def __call__(self, graph) -> bool:
        return self.simp(graph)

class RewriteSingleVertex(Rewrite[VT, ET]):
    applier: Callable[[BaseGraph[VT, ET], VT], bool]
    is_match: Callable[[BaseGraph[VT, ET], VT], bool]

    def __init__(self, is_match: Callable[[BaseGraph[VT, ET], VT], bool],
                 applier: Callable[[BaseGraph[VT, ET], VT], bool]) -> None:
        super().__init__()
        self.is_match = is_match
        self.applier = applier
        self.__doc__ = is_match.__doc__

    def apply(self, graph: BaseGraph[VT, ET], v: VT) -> bool:
        if self.is_match(graph, v):
            self.applier(graph, v)
            return True
        return False


class RewriteSimpSingleVertex(RewriteSingleVertex[VT, ET]):
    simp_match: Optional[Callable[[BaseGraph[VT, ET], VT], bool]]

    def __init__(self, is_match: Callable[[BaseGraph[VT, ET], VT], bool],
                 applier: Callable[[BaseGraph[VT, ET], VT], bool],
                 simp_match: Optional[Callable[[BaseGraph[VT, ET], VT], bool]] = None) -> None:
        super().__init__(is_match, applier)
        self.simp_match = simp_match

    def find_all_matches (self, graph: BaseGraph[VT, ET]) -> Set[VT]:
        all_matches: Set[VT] = set()
        if self.simp_match is not None:
            match = self.simp_match
        else:
            match = self.is_match

        for v in graph.vertices():
            if match(graph, v):
                all_matches.add(v)
        return all_matches

    def simp(self, graph: BaseGraph[VT, ET]) -> bool:
        if self.simp_match is not None:
            match = self.simp_match
        else:
            match = self.is_match
        applied: bool = False
        while True:
            j = 0
            all_matches = self.find_all_matches(graph)
            for m in all_matches:
                if match(graph, m):
                    j += 1
                    self.applier(graph, m)
                    applied = True
            if j == 0:
                break
        return applied

class RewriteDoubleVertex(Rewrite[VT, ET]):
    applier: Callable[[BaseGraph[VT, ET], VT, VT], bool]
    is_match: Callable[[BaseGraph[VT, ET], VT, VT], bool]

    def __init__(self, is_match: Callable[[BaseGraph[VT, ET], VT, VT], bool],
                 applier: Callable[[BaseGraph[VT, ET], VT, VT], bool]) -> None:
        super().__init__()
        self.is_match = is_match
        self.applier = applier
        self.__doc__ = is_match.__doc__

    def apply(self, graph: BaseGraph[VT, ET], v1: VT, v2: VT) -> bool:
        if self.is_match(graph, v1, v2):
            self.applier(graph, v1, v2)
            return True
        return False


class RewriteSimpDoubleVertex(RewriteDoubleVertex[VT, ET]):
    simp_match: Optional[Callable[[BaseGraph[VT, ET], VT, VT], bool]]

    def __init__(self, is_match: Callable[[BaseGraph[VT, ET], VT, VT], bool],
                 applier: Callable[[BaseGraph[VT, ET], VT, VT], bool],
                 simp_match: Optional[Callable[[BaseGraph[VT, ET], VT, VT], bool]] = None) -> None:
        super().__init__(is_match, applier)
        self.simp_match = simp_match

    def find_all_matches (self, graph: BaseGraph[VT, ET]) -> Set[Tuple[VT, VT]]:
        all_matches: Set[Tuple[VT, VT]] = set()
        if self.simp_match is not None:
            match = self.simp_match
        else:
            match = self.is_match

        for v1 in graph.vertices():
            for v2 in graph.neighbors(v1):
                if match(graph, v1, v2):
                    pair = (v1, v2) if v1 <= v2 else (v2, v1)
                    all_matches.add(pair)
        return all_matches

    def simp(self, graph: BaseGraph[VT, ET]) -> bool:
        if self.simp_match is not None:
            match = self.simp_match
        else:
            match = self.is_match

        applied: bool = False
        while True:
            j = 0
            all_matches = self.find_all_matches(graph)

            for m in all_matches:
                if match(graph, m[0], m[1]):
                    j += 1
                    self.applier(graph, m[0], m[1])
                    applied = True
            if j == 0:
                break
        return applied


from .utils import VertexType
from .graph import Graph
from .linalg import Mat2
from typing import Optional

def generate_css_encoder_graph(S: Mat2, L: Optional[Mat2]=None, type: str='Z-X'):
    """Returns a phase-free PyZX Graph of the encoder for a CSS code,
      given its stabilizers (S), logical operators (L), and normal form type (type).
      Normal form type can be 'Z-X' or 'X-Z' (Definitions 4.3.1 and 4.3.7 of Picturing Quantum Software).
      For type 'Z-X'('X-Z'), use X(Z)-type stabilizers and logicals.
      
      Example:
		To construct the encoder for the [[8, 3, 2]] code in Z-X form:

            SX = Mat2([[1,1,1,1,1,1,1,1]])
			LX = Mat2([[1,1,1,1,0,0,0,0],[1,1,0,0,1,1,0,0],[1,0,1,0,1,0,1,0]])
            enc, vertex_ids = generate_css_encoder_graph(SX, LX, 'Z-X')
      
    """
    
    if type == 'Z-X':
        r1type, r2type = VertexType.Z, VertexType.X
    elif type == 'X-Z':
        r1type, r2type = VertexType.X, VertexType.Z

    g = Graph()
    logical_verts = []
    stabilizer_verts = []
    output_verts = []
    num_L = L.rows() if L is not None else 0
    num_S = S.rows() if S is not None else 0
    n = S.cols() if S is not None else 0
    
    for i in range(num_L):
        in_bound_vert = g.add_vertex(VertexType.BOUNDARY, qubit=i, row=0)
        logical_vert = g.add_vertex(r1type, qubit=i, row=2)
        g.add_edge((in_bound_vert, logical_vert))
        logical_verts.append(logical_vert)

    for i in range(num_L, num_L+num_S):
        stabilizer_vert = g.add_vertex(r1type, qubit=i, row=2)
        stabilizer_verts.append(stabilizer_vert)
    
    for j in range(n):
        output_vert = g.add_vertex(r2type, qubit=j, row=6)
        out_bound_vert = g.add_vertex(VertexType.BOUNDARY, qubit=j, row=8)
        g.add_edge((output_vert, out_bound_vert))
        output_verts.append(output_vert)

    for i in range(num_L+num_S):
        for j in range(n):
            if (L is not None) and (i < num_L) and (L[i,j] == 1):
                g.add_edge((logical_verts[i], output_verts[j]))
            elif (i >= num_L) and (S[i-num_L,j] == 1):
                g.add_edge((stabilizer_verts[i-num_L], output_verts[j]))
    
    return g, [logical_verts, stabilizer_verts, output_verts]


# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module implements several optimization methods on ``Circuit``\\ s. 
The function :func:`basic_optimization` runs a set of back-and-forth gate commutation and cancellation routines.
:func:`phase_block_optimize` does phase polynomial optimization using the TODD algorithm,
and :func:`full_optimize` combines these two methods."""

from typing import overload, Tuple, List, Union, Dict, Set
from typing_extensions import Literal

from .circuit import Circuit
from .circuit.gates import Gate, ZPhase, XPhase, CNOT, CZ, ParityPhase, NOT, HAD, SWAP, S, Z
from .extract import permutation_as_swaps
from .todd import todd_simp

__all__ = ['full_optimize', 'basic_optimization', 'phase_block_optimize']

def full_optimize(circuit: Circuit, quiet:bool=True) -> Circuit:
    """Optimizes the circuit using first some basic commutation and cancellation rules,
    and then a dedicated phase polynomial optimization strategy involving the TODD algorithm.

    Args:
        circuit: Circuit to be optimized.
        quiet: Whether to print some progress indicators."""
    c = basic_optimization(circuit.to_basic_gates())
    c = phase_block_optimize(c, quiet=quiet)
    return basic_optimization(c.to_basic_gates())

def basic_optimization(circuit: Circuit, do_swaps:bool=True, quiet:bool=True) -> Circuit:
    """Optimizes the circuit using a strategy that involves delayed placement of gates
    so that more matches for gate cancellations are found. Specifically tries to minimize
    the number of Hadamard gates to improve the effectiveness 
    of phase-polynomial optimization techniques.

    Args:
        circuit: Circuit to be optimized.
        do_swaps: When set uses some rules transforming CNOT gates into SWAP gates. Generally leads to better results, but messes up architecture-aware placement of 2-qubit gates.
        quiet: Whether to print some progress indicators.
    """
    if not isinstance(circuit, Circuit):
        raise TypeError("Input must be a Circuit")
    o = Optimizer(circuit)
    return o.parse_circuit(do_swaps=do_swaps,quiet=quiet)

def toggle_element(l:list, e:object) -> None:
    if e in l: l.remove(e)
    else: l.append(e)

def swap_element(l:list, e1:object, e2:object) -> None:
    if e1 in l and e2 not in l:
        l.remove(e1)
        l.append(e2)
    elif e2 in l and e1 not in l:
        l.remove(e2)
        l.append(e1)

def stats(circ: Circuit) -> Tuple[int,int,int]:
    two_qubit = 0
    had = 0
    non_pauli = 0
    for g in circ.gates:
        if g.name in ('CZ', 'CNOT'):
            two_qubit += 1
        elif g.name == 'HAD':
            had += 1
        elif g.name != 'NOT' and g.phase != 1:
            non_pauli += 1
    return had, two_qubit, non_pauli

class Optimizer(object):
    """Helper class that implements the optimizations of :func:`basic_optimization`.
    Works by doing alternating forward and backward 'passes' through a circuit.
    During a pass, we iteratively consume gates, while 
    keeping track of a stack of mutually commuting gates on each qubit.
    When the gate can be combined with a gate on the stack, this is done.
    If it doesn't combine, but commutes with the gates on the stack, it is added to the stack.
    If it doesn't commute, the stack is reset."""
    def __init__(self, circuit: Circuit) -> None:
        self.circuit: Circuit = circuit
        self.qubits: int = circuit.qubits
        self.minimize_czs: bool = False
        self.do_swaps: bool = True
    
    @overload
    def parse_circuit(self, 
            separate_correction:Literal[False]=False, 
            max_iterations:int=1000, 
            do_swaps:bool=True, 
            quiet:bool=True) -> Circuit: pass

    @overload
    def parse_circuit(self, 
            separate_correction:Literal[True], 
            max_iterations:int=1000, 
            do_swaps:bool=True, 
            quiet:bool=True) -> Tuple[Circuit, List[Gate]]: pass

    def parse_circuit(self, 
            separate_correction:bool=False, 
            max_iterations:int=1000, 
            do_swaps:bool=True, 
            quiet:bool=True
            ) -> Union[Circuit, Tuple[Circuit, List[Gate]]]:
        """Repeatedly does forward and backward passes trough the circuit, 
        until no more improvements are found.
        When ``separate_correction`` is True, it returns the list of NOT, Z and SWAP gates that is 
        has collected at the end separately.

        When ``do_swaps`` is set to False, it does not apply the rule 
        CNOT(r1,r2)CNOT(r2,r1) = CNOT(r1,r2)SWAP, so that the connectivity of the circuit is preserved,
        in case the circuit already had been mapped."""
        self.do_swaps = do_swaps
        self.minimize_czs = False
        self.circuit, correction = self.parse_forward()
        count = stats(self.circuit)
        for g in correction: self.circuit.gates.extend(g.to_basic_gates())
        i = 0
        while True:
            self.circuit.gates = list(reversed(self.circuit.gates))
            self.circuit, correction = self.parse_forward()
            for g in correction: self.circuit.gates.extend(g.to_basic_gates())
            self.circuit.gates = list(reversed(self.circuit.gates))
            self.circuit, correction = self.parse_forward()
            i += 1
            s = stats(self.circuit)
            if self.minimize_czs and (all(s1<=s2 for s1,s2 in zip(count,s)) or i>=max_iterations): break
            for g in correction: self.circuit.gates.extend(g.to_basic_gates())
            if not quiet:
                print(i, end='.')
            count = s
            self.minimize_czs = True
        for g in self.circuit.gates: g.index = 0
        if not separate_correction:
            for g in correction: self.circuit.gates.extend(g.to_basic_gates())
            return self.circuit
        else:
            return self.circuit, correction
    
    def parse_forward(self) -> Tuple[Circuit, List[Gate]]:
        """Does a single forward pass trough self.circuit.gates."""
        self.gates: Dict[int,List[Gate]] = {i:list() for i in range(self.qubits)}
        self.available: Dict[int,List[Gate]] = {i:list() for i in range(self.qubits)}
        self.availty = {i: 1 for i in range(self.qubits)}
        self.hadamards: List[int] = []
        self.nots: List[int] = []
        self.zs: List[int] = []
        self.permutation = {i:i for i in range(self.qubits)}
        self.gcount = 0
        for g in self.circuit.gates:
            self.parse_gate(g)
        for t in self.hadamards.copy():
            self.add_hadamard(t)
        for t in self.zs:
            z = Z(t)
            z.index = self.gcount
            self.gcount += 1
            self.gates[t].append(z)
        # for t in self.nots:
        #     n = NOT(t)
        #     #correction.append(n)
        #     n.index = self.gcount
        #     self.gcount += 1
        #     self.gates[t].append(n)
        
        c = Circuit(self.qubits)
        c.gates = self.topological_sort_gates()
        
        correction: List[Gate] = []
        for t in self.nots:
            n = NOT(t)
            correction.append(n)
            # n.index = self.gcount
            # self.gcount += 1
            # self.gates[t].append(n)
        swaps = permutation_as_swaps(self.permutation)
        for a,b in swaps:
            correction.append(SWAP(a,b))
            #c.gates.extend(SWAP(a,b).to_basic_gates())
        return c, correction

    def topological_sort_gates(self) -> List[Gate]:
        """self.gates is a a {qubit:[list of gates]} dictionary. This function consumes this dictionary and outputs a
        single list of gates, with the gates in the correct order.
        Note that 2-qubit gates are present in two entries in the dictionary and are identified with an ``index`` parameter."""
        output = []
        while any(self.gates.values()):
            available_indices: Set[int] = set()
            for q, gs in self.gates.items():
                while gs:
                    g = gs[0]
                    if g.name not in ('CZ', 'CNOT'):
                        output.append(gs.pop(0))
                    elif g.index in available_indices:
                        available_indices.remove(g.index)
                        q2 = g.target if q == g.control else g.control
                        self.gates[q2].remove(g)
                        output.append(gs.pop(0))
                    else:
                        ty = 1 if (g.name == 'CZ' or g.control == q) else 2
                        available_indices.add(g.index)
                        remove = []
                        for i, g2 in enumerate(gs[1:]):
                            if (ty == 1 and isinstance(g2, ZPhase)) or (ty == 2 and isinstance(g2, XPhase)):
                                output.append(g2)
                                remove.append(i)
                            elif g2.name not in ('CZ', 'CNOT'): break
                            elif ((ty == 1 and (g2.name == 'CZ' or g2.control == q)) or
                                  (ty == 2 and g2.name == 'CNOT' and g2.target == q)):
                                if g2.index in available_indices:
                                    available_indices.remove(g2.index)
                                    q2 = g2.target if q == g2.control else g2.control
                                    self.gates[q2].remove(g2)
                                    output.append(g2)
                                    remove.append(i)
                                else:
                                    available_indices.add(g2.index)
                            else:
                                break
                        for i in reversed(remove):
                            gs.pop(i+1)
                        break
        return output

    
    def add_hadamard(self, t: int) -> None:
        """Called by ``parse_gate`` to add a Hadamard gate to the output."""
        h = HAD(t)
        h.index = self.gcount
        self.gcount += 1
        self.gates[t].append(h)
        self.hadamards.remove(t)
        self.available[t] = list()
        self.availty[t] = 1
    
    def add_gate(self, t: int, g: Gate) -> None:
        """Helper function for ``add_cz`` and ``add_cnot`` to add a single qubit gate to the output."""
        g.index = self.gcount
        self.gcount += 1
        self.gates[t].append(g)
        self.available[t].append(g)
    
    def add_cz(self, cz: CZ) -> None:
        """Called by ``parse_gate`` to add a CZ gate to the output.
        Does some non-trivial logic to see whether the CZ-gate can be cancelled against a CNOT or CZ gate."""
        t1, t2 = cz.control, cz.target
        #We first try to find a matching CNOT gate
        found_match = False
        if self.minimize_czs:
            for c,t in [(t1,t2),(t2,t1)]:
                for g in self.available[c]:
                    if g.name == 'CNOT' and g.control == c and g.target == t:
                        if self.availty[t] == 2:
                            if g in self.available[t]: # The gate is also available on the target qubit
                                found_match = True
                                break
                            else:
                                continue
                        # There are Z-like gates blocking the CNOT from usage
                        # But if the CNOT can be passed all the way up to these Z-like gates
                        # Then we can commute the CZ gate next to the CNOT and hence use it.
                        for h in list(reversed(self.gates[t][:-len(self.available[t])])): # We start looking at the gates behind the Z-like gates
                            if h.name != 'CNOT' or h.target != t: # If any of those gates is not a CNOT of the right type, then we stop our search
                                break
                            if h == g: # But if all the previous gates are fine, than we can use this CNOT.
                                found_match = True
                                break
                        if found_match: break
                if found_match: break
        if found_match: #CNOT-CZ = (S* x id)CNOT (S x S)
            t,c = g.target, g.control
            if self.availty[t] == 2:
                self.availty[t] == 1
                self.available[t] = []
            self.gates[t].remove(g)
            self.gates[c].remove(g)
            self.available[c].remove(g)
            s1 = S(t, adjoint=True)
            s1.index = self.gcount
            self.gcount += 1
            if self.available[t]: # There are gates of non-commuting type on this qubit
                self.gates[t].insert(-len(self.available[t]),s1) # And hence we must insert these gates at
                self.gates[t].insert(-len(self.available[t]),g)  # the correct location
            else: 
                self.gates[t].append(s1)
                self.gates[t].append(g)
            s2 = S(t)
            s2.index = self.gcount
            self.gcount += 1
            self.gates[t].append(s2) # In contrast, these gates appear after the CNOT, necessarily on Z-like phases
            self.available[t].append(s2) # and hence can be added at the end of the list
            s3 = S(c)
            s3.index = self.gcount
            self.gcount += 1
            self.available[c].append(g)
            self.available[c].append(s3)
            self.gates[c].append(g)
            self.gates[c].append(s3)
            return

        if self.availty[t1] == 2:
            self.available[t1] = list()
            self.availty[t1] = 1
        if self.availty[t2] == 2:
            self.available[t2] = list()
            self.availty[t2] = 1

        found_match = False
        for g in reversed(self.available[t1]): # We try to find a CZ with the same control and target
            if g.name == 'CZ' and g.control == t1 and g.target == t2: # Here it is important that we have normalised all CZs
                found_match = True                                    # to have cz.control<cz.target
                break
        if found_match:
            if g not in self.available[t2]: # We still need to check if the CZ is actually available on the other qubit
                found_match = False
            else:
                self.available[t1].remove(g)
                self.gates[t1].remove(g)
                self.available[t2].remove(g)
                self.gates[t2].remove(g)

        if not found_match: # No cancellation found, so we just add the gate
            cz.index = self.gcount
            self.gcount += 1
            self.gates[t1].append(cz)
            self.gates[t2].append(cz)
            self.available[t1].append(cz)
            self.available[t2].append(cz)
    
    def add_cnot(self, cnot: CNOT) -> None:
        """Called by ``parse_gate`` to parse a CNOT gate.
        Does some non-trivial logic to see whether the CNOT gate can be cancelled against another CNOT gate on the same qubits."""
        c, t = cnot.control, cnot.target
        if self.availty[c] == 2:
            if self.availty[t] == 1: # Try to find anti-match
                found_match = False
                for g in reversed(self.available[c]):
                    if g.name == 'CNOT' and g.control == t and g.target == c:
                        found_match = True
                        break
                if found_match and self.do_swaps: # We do the CNOT(t,c)CNOT(c,t) = CNOT(c,t)SWAP(c,t) commutation
                    if g in self.available[t]:
                        self.gates[c].remove(g)
                        self.gates[t].remove(g)
                        self.availty[c] = 1
                        self.availty[t] = 2
                        cnot.index = self.gcount
                        self.gcount += 1
                        self.gates[c].append(cnot)
                        self.gates[t].append(cnot)
                        self.available[c] = [cnot]
                        self.available[t] = [cnot]
                        a = self.permutation[c]
                        b = self.permutation[t]
                        self.permutation[c] = b
                        self.permutation[t] = a
                        swap_element(self.hadamards, t, c)
                        swap_element(self.nots, t, c)
                        swap_element(self.zs, t, c)
                        return
                
            self.available[c] = list()
            self.availty[c] = 1
        if self.availty[t] == 1:
            self.available[t] = list()
            self.availty[t] = 2
        found_match = False
        for g in reversed(self.available[c]):
            if g.name == 'CNOT' and g.control == c and g.target == t:
                found_match = True
                break
        if found_match: # We do CNOT(c,t)CNOT(c,t) = id
            if g not in self.available[t]:
                found_match = False
            else:
                self.available[c].remove(g)
                self.gates[c].remove(g)
                self.available[t].remove(g)
                self.gates[t].remove(g)
                
        if not found_match:
            cnot.index = self.gcount
            self.gcount += 1
            self.gates[c].append(cnot)
            self.gates[t].append(cnot)
            self.available[c].append(cnot)
            self.available[t].append(cnot)
    
    def parse_gate(self, g: Gate) -> None:
        """The main function of the optimization. It records whether a gate needs to be placed at the specified location
        'right now', or whether we can postpone the placement until hopefully it is cancelled against some future gate.
        Only supports ZPhase, HAD, CNOT and CZ gates. """
        g = g.copy()
        # If we have some SWAPs recorded we need to change the target/control of the gate accordingly
        g.target = next(i for i in self.permutation if self.permutation[i] == g.target)
        t = g.target
        if g.name in ('CZ', 'CNOT'):
            g.control = next(i for i in self.permutation if self.permutation[i] == g.control)

        if g.name == 'HAD':
            # If we have recorded a NOT or Z gate at the target location, we push it trough the Hadamard and change the type
            if t in self.nots and t not in self.zs:
                self.nots.remove(t)
                self.zs.append(t)
            elif t in self.zs and t not in self.nots:
                self.zs.remove(t)
                self.nots.append(t)
            # See whether we have a HAD-S-HAD situation
            # And turn it into a S*-HAD-S* situation
            if len(self.gates[t])>1 and self.gates[t][-2].name == 'HAD' and isinstance(self.gates[t][-1], ZPhase):
                    g2 = self.gates[t][-1]
                    if g2.phase.denominator == 2:
                        h = self.gates[t][-2]
                        zp = ZPhase(t, (-g2.phase)%2)
                        zp.index = self.gcount
                        self.gcount += 1
                        g2.phase = zp.phase
                        if g2.name == 'S' and g2.phase.numerator > 1:
                            g2.adjoint = True
                        self.gates[t].insert(-2,zp)
                        return
            toggle_element(self.hadamards, t)
        elif g.name == 'NOT':
            toggle_element(self.nots, t)
        elif isinstance(g, ZPhase):
            if t in self.zs: #Consume a Z gate into the phase gate
                g.phase = (g.phase+1)%2
                self.zs.remove(t)
            if g.phase == 0: return
            if t in self.nots: # Push the phase gate trough a NOT
                g.phase = (-g.phase)%2
            if g.phase == 1: # If the resulting phase is a pi, then we record it as a Z gate
                toggle_element(self.zs, t)
                return
            if g.name == 'S':                           # We might have changed the phase, and therefore
                g.adjoint = g.phase.numerator != 1      # Need to adjust whether the adjoint is true
            if t in self.hadamards: # We can't push a phase gate trough a HAD, so we actually place the HAD down
                self.add_hadamard(t)
            if self.availty[t] == 1 and any(isinstance(g2, ZPhase) for g2 in self.available[t]): # There is an available phase gate
                i = next(i for i,g2 in enumerate(self.available[t]) if isinstance(g2, ZPhase))   # That we can fuse with the new one
                g2 = self.available[t].pop(i)
                self.gates[t].remove(g2)
                phase = (g.phase+g2.phase)%2
                if phase == 1:
                    toggle_element(self.zs, t)
                    return
                if phase != 0:
                    p = ZPhase(t, phase)
                    self.add_gate(t,p)
            else:
                if self.availty[t] == 2: # If previous gate was of X-type
                    self.availty[t] = 1  # We reset the available gates on this qubit
                    self.available[t] = list()
                g = ZPhase(t, g.phase)  # Avoid subclasses of ZPhase with inconsistent phase
                self.add_gate(t, g)
        elif g.name == 'CZ':
            t1, t2 = g.control, g.target
            if t1 > t2: # Normalise so that always g.target<g.control (since CZs are symmetric anyway)
                g.target = t1
                g.control = t2
            # Push NOT gates trough the CZ
            if t1 in self.nots: 
                toggle_element(self.zs, t2)
            if t2 in self.nots:
                toggle_element(self.zs, t1)
            # If there are HADs on both targets, we cannot commute the CZ trough and we place the HADs
            if t1 in self.hadamards and t2 in self.hadamards:
                self.add_hadamard(t1)
                self.add_hadamard(t2)
            if t1 not in self.hadamards and t2 not in self.hadamards:
                self.add_cz(g)
            # Exactly one of t1 and t2 has a hadamard
            # So the CZ commutes trough and becomes a CNOT
            elif t1 in self.hadamards:
                cnot = CNOT(t2, t1)
                self.add_cnot(cnot)
            else:
                cnot = CNOT(t1, t2)
                self.add_cnot(cnot)
            
        elif g.name == 'CNOT':
            c, t = g.control, g.target
            # Commute NOTs and Zs trough the CNOT
            if c in self.nots:
                toggle_element(self.nots, t)
            if t in self.zs:
                toggle_element(self.zs, c)
            # If HADs are on both qubits, we commute the CNOT trough by switching target and control
            if c in self.hadamards and t in self.hadamards:
                g.control = t
                g.target = c
                self.add_cnot(g)
            elif c not in self.hadamards and t not in self.hadamards:
                self.add_cnot(g)
            # If there is a HAD on the target, the CNOT commutes trough to become a CZ
            elif t in self.hadamards:
                cz = CZ(c if c<t else t, c if c>t else t)
                self.add_cz(cz)
            else: # Only the control has a hadamard gate in front of it
                self.add_hadamard(c)
                self.add_cnot(g)

        else:
            raise TypeError("Unknown gate {}. Maybe simplify the gates with circuit.to_basic_gates()?".format(str(g)))




def greedy_consume_gates(gates: Dict[int, List[Gate]], qubits: int) -> Tuple[List[Gate],List[HAD]]:
    """Tries to consume as many gates as possible into a phase-polynomial block, by pushing gates past hadamards to the beginning
    as long as that is possible. Used by :func:`phase_block_optimize`.

    ``gates`` should be a {qubits:[list of gates]} dictionary, while ``qubits`` is the amount of qubits in the circuit.
    Returns a tuple (list of gates, list of hadamards)."""
    
    block = [] # The output
    while True:
        had_blocked = dict() # a {qubit: HADgate} dictionary specifying when a HAD blocks further consuming of gates.
        to_be_appended = [] # List of gates that we will add to ``block``.
        available = []      # List of indices of 2-qubit gates to record whether they are available to be added on the other target.
        gatetype = {i: 0 for i in range(qubits)} # 0 = Z-type, 1 = X-type, the two sorts of commutation types.
        for q, gs in gates.items():
            if not gs: continue
            g = gs[0]
            if g.name == 'HAD': # If the first gate on this qubit is a HAD, we stop
                had_blocked[q] = g
                continue
            if isinstance(g, ZPhase) or g.name == 'CZ':
                gatetype[q] = 1
            else: # gate is a CNOT
                if g.control == q:
                    gatetype[q] = 1
                else:
                    gatetype[q] = 2
            for g in gs:
                if g.name == 'HAD': # Stop once we encounter a HAD
                    had_blocked[q] = g
                    break
                if isinstance(g, ZPhase) or g.name == 'CZ': # Z-type gates
                    if gatetype[q] == 1: # Z-type is available
                        if g.name == 'CZ':
                            if g.index in available: # Check whether the target on the other qubit is available
                                to_be_appended.append(g)
                            else: available.append(g.index) # Otherwise we postpone until we have checked that later on
                        else:
                            to_be_appended.append(g)
                    else:
                        break # We have encountered a gate of the wrong type, so we stop delving deeper
                else: #gate is CNOT
                    if (gatetype[q] == 1 and g.target == q) or (gatetype[q] == 2 and g.control == q): # wrong type
                        break
                    else:
                        if g.index in available: # Same 2-qubit gate logic as with CZ
                            to_be_appended.append(g)
                        else: available.append(g.index)
        for g in to_be_appended:
            block.append(g)
            gates[g.target].remove(g)
            if g.name in ('CZ', 'CNOT'):
                gates[g.control].remove(g)  
        if to_be_appended: # We added at least one gate, so we go to the top of the loop to try again.
            continue
        # We couldn't add any easy gates, so now we go looking for gates stuck behind a HAD.
        added_any = False
        candidates = []
        for q, had in had_blocked.items():
            i = gates[q].index(had)
            gs = gates[q][i+1:] # The gates appearing after the HAD
            if not gs: continue
            g = gs[0]
            if g.name == 'HAD': # Double Hadamard
                gates[q].remove(had)
                gates[q].remove(g)
                added_any = True
                break
            left_ty = gatetype[q] # The type of the gates to the left of the HAD. Note that this type must necessarily
                                  #  be the same for all gates, since otherwise it wouldn't be blocked by a HAD.
            if g.name == 'CZ' or isinstance(g, ZPhase) or (g.control == q):
                if gatetype[q] == 0: left_ty = 2 # If no gate is on the left of the HAD we set the type correspondingly.
                right_ty = 1
            else: 
                if gatetype[q] == 0: left_ty = 1
                right_ty = 2
            if left_ty == right_ty: continue # If the types are different, we can't commute things past the HAD into the phase-block.
            for g in gs:
                if g.name == 'HAD': break # If we encounter another HAD, we stop.
                if isinstance(g, ZPhase):
                    if right_ty == 1: continue # We can't commute a ZPhase past a HAD, but we can keep looking further
                    else: break # ZPhase is not of type X, so we must stop looking now.
                if g.name == 'CZ' or g.control == q: # CZ or CNOT with a control on this qubit
                    if right_ty == 2: break
                else:  # CNOT with target on this qubit
                    if right_ty == 1: break
                if g.index not in available:
                    if g.name == 'CNOT':  # We only need to check CNOTs, since CZs must already be in available 
                        available.append(g.index)  # (because otherwise they would be behind 2 HADs)
                else:
                    if g not in candidates:
                        candidates.append(g)
        if added_any: continue # Found double Hadamard

        for g in candidates:
            if g.name == 'CZ':
                if g.target in had_blocked and g.index > had_blocked[g.target].index: # CZ appears after the HAD.
                    q = g.target
                else:
                    q = g.control
                q2 = g.target if g.control == q else g.control
                if q2 in had_blocked and g.index > had_blocked[q2].index:
                    print(g, g.index)
                    raise Exception("CZ behind two Hadamard gates. This is not supposed to happen")
                cnot = CNOT(q2, q)
                cnot.index = g.index
                gates[q].remove(g)
                gates[q2].remove(g)
                block.append(cnot)
                added_any = True
            elif g.name == 'CNOT':
                if (g.target in had_blocked and g.index > had_blocked[g.target].index
                     and g.control in had_blocked and g.index > had_blocked[g.control].index):
                    cnot = CNOT(g.target, g.control)
                    cnot.index = g.index
                    gates[g.target].remove(g)
                    gates[g.control].remove(g)
                    block.append(cnot)
                    added_any = True
                elif g.target in had_blocked and g.index > had_blocked[g.target].index:
                    cz = CZ(g.control, g.target)
                    cz.index = g.index
                    gates[g.target].remove(g)
                    gates[g.control].remove(g)
                    block.append(cz)
                    added_any = True
                else: continue
        if added_any: continue
        else: break

    hadamards = []
    for gs in gates.values():
        if gs and gs[0].name == 'HAD':
            hadamards.append(gs.pop(0))
    return block, hadamards


def phase_block_optimize(circuit: Circuit, pre_optimize:bool=True, quiet:bool=True) -> Circuit:
    """Optimizes the given circuit, by cutting it into phase polynomial pieces, and
    using the `TODD algorithm <https://iopscience.iop.org/article/10.1088/2058-9565/aad604/meta>`_ 
    to optimize each of these phase polynomials. 
    The phase-polynomial circuits are then resynthesized using the 
    `parity network <https://iopscience.iop.org/article/10.1088/2058-9565/aad8ca/meta>`_ algorithm.
    
    Note:
        Only works with Clifford+T circuits. Will give wrong output when fed smaller rotation gates, or Toffoli-like gates.
        Depending on the number of qubits and T-gates this function can take a long time to run.
        It can be sped up somewhat by using the `TOpt implementation of TODD <https://github.com/Luke-Heyfron/TOpt>`_.
        If this is installed, point towards it using ``zx.settings.topt_command``, such as for instance 
        ``zx.settings.topt_command = ['wsl', '../TOpt']`` for running it in the Windows Subsystem for Linux.

    Args:
        circuit: The circuit to be optimized.
        pre_optimize: Whether to call :func:`basic_optimization` first.
        quiet: Whether to print some progress indicators. Helpful when execution time is long. 
    """
    qubits = circuit.qubits
    o = Optimizer(circuit)
    if pre_optimize:
        circuit, correction = o.parse_circuit(separate_correction=True, quiet=quiet)
    else:
        circuit = circuit.copy()
        correction = []
    permutation = {i:i for i in range(qubits)}
    nots = []
    for g in correction:
        if g.name == 'SWAP':
            a = permutation[g.control]
            b = permutation[g.target]
            permutation[g.control] = b
            permutation[g.target] = a
        elif g.name == 'NOT':
            nots.append(g)
        else:
            raise TypeError("Illegal correction {}".format(str(g)))

    permutation = {v:k for k,v in permutation.items()}

    gates = {i:list() for i in range(qubits)}

    for i, g in enumerate(circuit.gates):
        g = g.copy()
        g.index = i
        if g.name in ('CNOT', 'CZ'):
            gates[g.control].append(g)
            gates[g.target].append(g)
        elif g.name != 'HAD':
            if not isinstance(g, ZPhase):
                raise TypeError("Unknown gate {}. Maybe simplify the gates with circuit.to_basic_gates()?".format(str(g)))
            elif g.phase.denominator not in (1,2,4):
                raise TypeError("This method only works on Clifford+T circuits. This circuit contains a {}. For these circuits, stick to basic_optimization().".format(str(g)))
            gates[g.target].append(g)
        else:
            gates[g.target].append(g)

    consumed = []

    block = []
    hadamards = []
    while any(gates.values()):
        if not quiet: print("new block")
        revgates = {i:list() for i in range(qubits)}
        i = 0
        for g in hadamards:
            g.index = i
            i += 1
            revgates[g.target].append(g)
        for g in reversed(block):
            g.index = i
            i += 1
            revgates[g.target].append(g)
            if g.name in ('CNOT', 'CZ'):
                revgates[g.control].append(g)

        revblock, had2 = greedy_consume_gates(revgates, qubits)
        if len(hadamards) != len(had2):
            raise Exception("Hadamards did not extract correctly")

        notparsed = []
        indices = set()
        for gs in revgates.values():
            for g in gs:
                if g.index not in indices:
                    indices.add(g.index)
                    notparsed.append(g)
        notparsed.sort(key=lambda g: g.index, reverse=True)

        consumed.extend(notparsed)
        consumed.extend(had2)

        newblock, hadamards = greedy_consume_gates(gates, qubits)
        block = list(reversed(revblock))
        block.extend(newblock)
        block, permute = todd_simp(block, qubits, quiet=quiet)
        inverse = {v:k for k,v in permute.items()}
        gates = {inverse[t]:gs for t,gs in gates.items()}
        indices = set()
        for gs in gates.values():
            for g in gs:
                if g.name in ('CNOT', 'CZ'):
                    if g.index in indices:
                        continue
                    indices.add(g.index)
                    g.control = inverse[g.control]
                g.target = inverse[g.target]
        for g in nots:
            g.target = inverse[g.target]
        permutation = {i: permutation[permute[i]] for i in range(qubits)}

    consumed.extend(block)
    consumed.extend(hadamards)
    consumed.extend(nots)
    for a,b in permutation_as_swaps(permutation):
        consumed.append(SWAP(a,b))
    for g in consumed: g.index = 0
    c = Circuit(qubits)
    c.gates = consumed
    return c



# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Tuple, Optional

from pyzx_param.routing.parity_maps import CNOT_tracker

from .architecture import Architecture
from ..linalg import Mat2

debug = False


def steiner_gauss(
    matrix: Mat2,
    architecture: Architecture,
    full_reduce: bool = False,
    x: Optional[CNOT_tracker] = None,
    y: Optional[CNOT_tracker] = None,
):
    """
    Performs Gaussian elimination that is constrained by the given architecture

    :param matrix: PyZX Mat2 matrix to be reduced
    :param architecture: The Architecture object to conform to
    :param full_reduce: Whether to fully reduce or only create an upper triangular form
    :param x: Optional CNOT_tracker object to track row operations
    :param y: Optional CNOT_tracker object to track column operations
    :return: Rank of the given matrix
    """

    def row_add(c0, c1):
        """
        Adds row c0 to c1 in the main matrix. If debug flag is set it prints what happening.

        :param c0: Source row
        :param c1: Target row
        """
        matrix.row_add(c0, c1)
        if debug:
            print("Reducing", c0, c1)
        if x != None:
            x.row_add(c0, c1)
        if y != None:
            y.col_add(c1, c0)

    def steiner_reduce(col: int, root: int, nodes: List[int], upper: bool):
        """
        Uses Steiner tree to reduce matrix columns

        :param col: Column to minimise
        :param root: Start point
        :param nodes: Target point
        :param upper: If true working on upper triangular part
        """
        steiner_tree = architecture.steiner_tree(root, nodes, upper)
        # Remove all zeros
        next_check = next(steiner_tree)
        if upper:
            zeros = []
            while next_check is not None:
                s0, s1 = next_check
                if matrix[s0, col] == 0:  # s1 is a new steiner point or root = 0
                    zeros.append(next_check)
                next_check = next(steiner_tree)
            while len(zeros) > 0:
                s0, s1 = zeros.pop(-1)
                if matrix[s0, col] == 0:
                    row_add(s1, s0)
                    if debug:
                        print(matrix[s0, col], matrix[s1, col])
            if (
                next_check is not None and matrix[next_check[0], col] == 0
            ):  # root is zero
                print("WARNING : Root is 0 => reducing non-pivot column", matrix.data)
            if debug:
                print("Step 1: remove zeros", matrix[:, col])
            while next_check is not None:
                s0, s1 = next_check
                if matrix[s1, col] == 0:  # s1 is a new steiner point
                    row_add(s0, s1)
                next_check = next(steiner_tree)
        # Reduce stuff
        if debug:
            print("Step 2: remove ones")
        next_add = next(steiner_tree)
        while next_add is not None:
            s0, s1 = next_add
            row_add(s0, s1)
            next_add = next(steiner_tree)
            if debug:
                print(next_add)
        if debug:
            print("Step 3: profit")

    rows = matrix.rows()
    cols = matrix.cols()
    p_cols = []
    pivot = 0
    current_row = 0

    for current_row in range(rows):
        found_pivot = False
        nodes = []
        while not found_pivot and pivot < cols:
            nodes = [r for r in range(current_row, rows) if matrix[r, pivot] == 1]
            if len(nodes) > 0:
                p_cols.append(pivot)
                found_pivot = True
            else:
                pivot += 1
        # no more pivots left
        if not found_pivot:
            break

        if nodes[0] != current_row:
            nodes.insert(0, current_row)
        steiner_reduce(pivot, current_row, nodes, True)
        pivot += 1

    # for c in range(cols):
    #     nodes = [r for r in range(pivot, rows) if pivot==r or matrix[r,c] == 1]
    #     steiner_reduce(c, pivot, nodes, True)
    #     if matrix[pivot,c] == 1:
    #         p_cols.append(c)
    #         pivot += 1

    if debug:
        print("Upper triangle form", matrix)
    rank = len(p_cols)
    if debug:
        print(p_cols)

    if full_reduce:
        for current_row in reversed(range(len(p_cols))):
            pivot = p_cols[current_row]
            nodes = [r for r in range(0, current_row) if matrix[r, pivot] == 1]
            if len(nodes) > 0:
                nodes.append(current_row)
                steiner_reduce(pivot, current_row, nodes, False)

    # if full_reduce:
    #     pivot -= 1
    #     for c in reversed(p_cols):
    #         debug and print(pivot, matrix[:,c])
    #         nodes = [r for r in range(0, pivot+1) if r==pivot or matrix[r,c] == 1]
    #         if len(nodes) > 1:
    #             steiner_reduce(c, pivot, nodes, False)
    #         pivot -= 1
    return rank


def rec_steiner_gauss(
    matrix: Mat2,
    architecture: Architecture,
    full_reduce: bool = False,
    x: Optional[CNOT_tracker] = None,
    y: Optional[CNOT_tracker] = None,
    permutation: Optional[List[int]] = None,
    **kwargs,
):
    """
    Performs Gaussian elimination that is constrained bij the given architecture according to https://arxiv.org/pdf/1904.00633.pdf
    Only works on full rank, square matrices.

    :param matrix: PyZX Mat2 matrix to be reduced
    :param architecture: The Architecture object to conform to
    :param full_reduce: Whether to fully reduce or only create an upper triangular form
    :param x: Optional CNOT_tracker object to track row operations
    :param y: Optional CNOT_tracker object to track column operations
    :param permutation: Optional permutation of the qubits
    """
    if permutation is None:
        permutation = [i for i in range(matrix.rows())]
    else:
        matrix.permute_cols(permutation)

    def row_add(c0, c1):
        """
        Adds row c0 to c1 in the main matrix. If debug flag is set it prints what happening.

        :param c0: Source row
        :param c1: Target row
        """
        matrix.row_add(c0, c1)
        if debug:
            print("Reducing", c0, c1)
        if x != None:
            x.row_add(c0, c1)
        if y != None:
            y.col_add(c1, c0)

    def steiner_reduce(col, root, nodes, usable_nodes, rec_nodes, upper):
        """
        Uses Steiner tree to reduce matrix columns

        :param col: Column to minimise
        :param root: Start point
        :param nodes: Target point
        :param upper: If true working on upper triangular part
        """
        if not all([q in usable_nodes for q in nodes]):
            raise Exception(
                "Terminals not in the subgraph "
                + str(upper)
                + str((col, root, nodes, usable_nodes, rec_nodes))
            )
        generator = steiner_reduce_column(
            architecture,
            [row[col] for row in matrix.data],
            root,
            nodes,
            usable_nodes,
            rec_nodes,
            upper,
        )
        cnot = next(generator, None)
        tree_nodes = []
        while cnot is not None:
            if (
                cnot[0] not in usable_nodes + rec_nodes
                or cnot[1] not in usable_nodes + rec_nodes
            ):
                raise Exception("Steiner tree not sticking to constraints")
            tree_nodes.extend(cnot)
            row_add(*cnot)
            cnot = next(generator, None)
        return tree_nodes

    def rec_step(qubit_removal_order):
        """
        Recursive step function to reduce matrix

        :param qubit_removal_order: List of row/column indices to work on
        """

        # size, p_cols and pivot is needed if the matrix isn't square or of full rank
        size = len(qubit_removal_order)
        # Order the rows and columns to be ascending.
        pivots = list(sorted(qubit_removal_order))
        for pivot_idx, c in enumerate(pivots):
            # Get all the nodes in the row below the diagonal (rows[i:]) where the entry equals 1 or it is the diagonal
            nodes = [r for r in pivots[pivot_idx:] if c == r or matrix.data[r][c] == 1]
            # Perform the steiner_reduce on the current column (c) with root rows2[pivot] and you can use the full matrix
            steiner_reduce(c, c, nodes, pivots[pivot_idx:], [], True)
        # Quick check upper triangular form - should never be printing!
        if not all(
            [all([matrix.data[i][j] == 0 for j in range(0, i)]) for i in range(size)]
        ):
            print(
                "This should never be printed. If you read this, there is a bug in pyzx/routing/steiner.py"
            )
            print()
            print("not upper triangular form?")
            for row in matrix.data:
                print(row)
            print("--------")
        # Full reduce requires the recursion
        if full_reduce:
            # We precalculated the maximal leafs in R (in the qubit_removal_order)
            for i, c in enumerate(qubit_removal_order):
                # Vertices we can still use steiner trees
                usable_nodes = qubit_removal_order[i:]  # = R

                # Pick the maximal vertex k in R: k = max(usable_nodes)
                # Pick the maximal leaf k' in R: k' = cols[i] = c
                # Let W be the set of vertices in the shortest path from k' to k (inclusive)
                path = architecture.shortest_path(
                    c, max(usable_nodes)
                )  # , usable_nodes)
                if path is None:
                    raise ValueError(
                        f"The architecture is not connected using when considering nodes 0..{max(usable_nodes)}"
                    )
                rec_qubits = [architecture.vertex2qubit(v) for v in path]

                # Apply steiner up:
                # Pick the nodes of the steiner tree: all rows with a 1 in column c or the pivot row (diagonal)
                nodes = [r for r in usable_nodes if (r == c or matrix.data[r][c] == 1)]

                if (
                    len(nodes) > 1
                ):  # Otherwise it is the diagonal, which is trivial/done
                    steiner_reduce(c, c, nodes, usable_nodes, rec_qubits, False)

                # Do recursion on the recursive nodes that were allowed to break the the upper triangular form.
                if len(rec_qubits) > 1:  # Trivial otherwise
                    rec_step(
                        list(reversed(rec_qubits))
                    )  # [n for n in rows if n in rec_qubits])

    # The qubit order is the order in which the spanning tree R is traversed
    rec_step(architecture.reduce_order)


def steiner_reduce_column(
    architecture: Architecture,
    col: List[int],
    root: int,
    nodes: List[int],
    usable_nodes: List[int],
    rec_nodes,
    upper: bool,
):
    """
    Performs Steiner tree reduction to a matrix column under the constraints of some quantum architecture

    :param architecture: The architecture graph to be used
    :param col: Column to minimise
    :param root: Start point
    :param nodes: Target point
    :param usable_nodes: A list of all possible nodes that can be used
    :param rec_nodes: A list of all nodes that must be used
    :param upper: If true working on upper triangular part
    :yeilds: Showcases which row additions are applied
    """
    steiner_tree = architecture.rec_steiner_tree(
        root, nodes, usable_nodes, rec_nodes, upper
    )
    # Remove all zeros
    next_check = next(steiner_tree)
    if debug:
        print("Step 1: remove zeros")
    if upper:
        zeros = []
        while next_check is not None:
            s0, s1 = next_check
            if col[s0] == 0:  # s1 is a new steiner point or root = 0
                zeros.append(next_check)
            next_check = next(steiner_tree)
        while len(zeros) > 0:
            s0, s1 = zeros.pop(-1)
            if col[s0] == 0:
                col[s0] = (col[s1] + col[s0]) % 2
                yield s1, s0
                if debug:
                    print(col[s0], col[s1])
    else:
        if debug:
            print("deal with zero root")
        if next_check is not None and col[next_check[0]] == 0:  # root is zero
            print("WARNING : Root is 0 => reducing non-pivot column", col)
        if debug:
            print("Step 1: remove zeros", col)
        while next_check is not None:
            s0, s1 = next_check
            if col[s1] == 0:  # s1 is a new steiner point
                col[s1] = (col[s1] + col[s0]) % 2
                yield s0, s1
            next_check = next(steiner_tree)
    # Reduce stuff
    if debug:
        print("Step 2: remove ones")
    next_add = next(steiner_tree)
    while next_add is not None:
        s0, s1 = next_add
        col[s1] = (col[s1] + col[s0]) % 2
        yield s0, s1
        next_add = next(steiner_tree)
        if debug:
            print(next_add)
    if debug:
        print("Step 3: profit")



# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fractions import Fraction
from typing import Dict, Set, Tuple, Optional

from .linalg import Mat2
from .graph.base import BaseGraph, VT, ET
from .utils import vertex_is_zx


def gflow(
    g: BaseGraph[VT, ET], focus: bool=False, reverse: bool=False, pauli: bool=False
) -> Optional[Tuple[Dict[VT, int], Dict[VT, Set[VT]]]]:
    r"""Compute the gflow of a diagram in graph-like form.

    :param g: A graphlike ZX diagram.
    :param focus: Compute the focussed gflow
    :param reverse: Reverse the roles of inputs and outputs
    :param pauli: Compute the Pauli flow, restricted to {XY, X, Y} measurements

    Based on algorithm by Perdrix and Mhalla.
    See dx.doi.org/10.1007/978-3-540-70575-8_70

    Slightly extended to allow searching for Pauli flow with measurement planes {XY, X, Y}.

    Here is the pseudocode it is based on:
    ```
    input : An open graph
    output: A generalised flow

    gFlow (V,Gamma,In,Out) =
    begin
      for all v in Out do
        l(v) := 0
      end
      return gFlowaux (V,Gamma,In,Out,1)
    end

    gFlowaux (V,Gamma,In,Out,k) =
    begin
      C := {}
      for all u in V \\ Out do
        Solve in F2 : Gamma[V \\ Out, Out \\ In] * I[X] = I[{u}]
        if there is a solution X0 then
          C := C union {u}
          g(u) := X0
          l(u) := k
        end
      end
      if C = {} then
        return (Out = V,(g,l))
      else
        return gFlowaux (V, Gamma, In, Out union C, k + 1)
      end
    end
    ```
    """
    l: Dict[VT, int] = {}
    gflow: Dict[VT, Set[VT]] = {}
    ty = g.types()

    vertices: Set[VT] = set(v for v in g.vertices() if vertex_is_zx(ty[v]))
    pattern_inputs: Set[VT] = set()
    pattern_outputs: Set[VT] = set()
    pauli_x: Set[VT] = set()
    pauli_y: Set[VT] = set()

    for inp in g.inputs():
        pattern_inputs |= set(n for n in g.neighbors(inp) if vertex_is_zx(ty[n]))
    for outp in g.outputs():
        pattern_outputs |= set(n for n in g.neighbors(outp) if vertex_is_zx(ty[n]))
    
    if reverse:
        pattern_inputs, pattern_outputs = pattern_outputs, pattern_inputs

    if pauli:
        for v in vertices:
            p = g.phase(v) % 2
            if p in (0,1):
                pauli_x.add(v)
            elif p in (Fraction(1,2), Fraction(3,2)):
                pauli_y.add(v)

    processed: Set[VT] = pattern_outputs.copy() | g.grounds()
    non_outputs = list(vertices.difference(pattern_outputs))
    zerovec = Mat2.zeros(len(non_outputs), 1)
    for v in processed:
        l[v] = 0

    k: int = 1
    while True:
        correct: Set[VT] = set()

        # a list of nodes that can currently be used in the correction set of the next node
        candidates = [
            v
            for v in (processed | pauli_x | pauli_y).difference(pattern_inputs)
            if focus or any(w not in processed for w in g.neighbors(v))
        ]

        if focus:
            clean = non_outputs
        else:
            clean = [v for v in vertices
                        if v not in processed and 
                        any(w in candidates for w in g.neighbors(v))]

        # compute the "flow-demand matrix", which is essentially the bi-adjacency matrix from
        # "clean" to "candidates", which additionally relates every Y-measured node to
        # itself.
        m = Mat2([[1 if g.connected(v,w) or (v==w and v in pauli_y) else 0 
                   for v in candidates] for w in clean])

        for index, u in enumerate(clean):
            if not focus or (u not in processed and any(w in candidates for w in g.neighbors(u))):
                vu = zerovec.copy()
                vu.data[index][0] = 1
                x = m.solve(vu)
                if x:
                    correct.add(u)
                    gflow[u] = {candidates[i] for i in range(x.rows()) if x.data[i][0]}
                    l[u] = k

        if not correct:
            if len(vertices) == len(processed):
                return {v: i if reverse else k - i - 1 for v,i in l.items()}, gflow
            return None
        else:
            processed.update(correct)
            k += 1




# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import random
import sys
from types import ModuleType
from typing import Optional

if __name__ == '__main__':
    sys.path.append('..')
    sys.path.append('.')
from pyzx_param.graph import Graph
from pyzx_param.graph.multigraph import Multigraph
from pyzx_param.generate import cliffords
from pyzx_param.circuit import Circuit

np: Optional[ModuleType]
try:
    import numpy as np
    from pyzx_param.tensor import tensorfy, compare_tensors, compose_tensors, adjoint, VertexType
except ImportError:
    np = None

SEED = 1337


@unittest.skipUnless(np, "numpy needs to be installed for this to run")
class TestTensor(unittest.TestCase):

    def test_scalar_difference(self):
        array = np.array([[0,1],[1,0]])
        scalar = 0.75 + 3j
        self.assertFalse(compare_tensors(scalar*array,array,True))
    def test_scalar_difference_ignore(self):
        array = np.array([[0,1],[1,0]])
        scalar = 0.75 + 3j
        self.assertTrue(compare_tensors(scalar*array,array,False))

    def test_trivial_inequality(self):
        array = np.array([[1,0],[0,1]])
        array2= np.array([[0,1],[1,0]])
        self.assertFalse(compare_tensors(array, array2))

    def test_id_graph(self):
        g = Graph()
        i = g.add_vertex(0,0,0)
        o = g.add_vertex(0,0,1)
        g.set_inputs((i,))
        g.set_outputs((o,))
        g.add_edge((i,o))
        t = tensorfy(g)
        id_array = np.array([[1,0],[0,1]])
        self.assertTrue(np.allclose(t,id_array))
        self.assertTrue(compare_tensors(t,id_array))

    def test_equality_of_id_zx_graph_to_id(self):
        g = Graph()
        i = g.add_vertex(0,0,0)
        o = g.add_vertex(0,0,2)
        g.set_inputs((i,))
        g.set_outputs((o,))
        g2 = g.copy()
        g.add_edge((i,o))
        v = g2.add_vertex(1,0,1)
        g2.add_edges([(i,v),(v,o)])
        tensor1 = tensorfy(g)
        tensor2 = tensorfy(g2)
        self.assertTrue(compare_tensors(tensor1,tensor2))

    def test_inequality_id_and_swap(self):
        g = Graph()
        i1 = g.add_vertex(0,0,0)
        i2 = g.add_vertex(0,1,0)
        o1 = g.add_vertex(0,0,1)
        o2 = g.add_vertex(0,1,1)
        g.set_inputs((i1, i2))
        g.set_outputs((o1, o2))
        g2 = g.copy()
        g.add_edges([(i1,o2),(i2,o1)])
        g2.add_edges([(i1,o1),(i2,o2)])
        id_id = tensorfy(g2)
        swap = tensorfy(g)
        self.assertFalse(compare_tensors(id_id,swap))

    def test_three_cnots_is_swap(self):
        g = Graph()
        i1 = g.add_vertex(0,0,0)
        i2 = g.add_vertex(0,1,0)
        o1 = g.add_vertex(0,0,1)
        o2 = g.add_vertex(0,1,1)
        g.set_inputs((i1, i2))
        g.set_outputs((o1, o2))
        g.add_edges([(i1,o2),(i2,o1)])
        swap = tensorfy(g)
        c = Circuit(2)
        c.add_gate("CNOT",0,1)
        c.add_gate("CNOT",1,0)
        c.add_gate("CNOT",0,1)
        three_cnots = tensorfy(c.to_graph())
        self.assertTrue(compare_tensors(swap,three_cnots))

    def test_compose(self):
        random.seed(SEED)
        circ1 = cliffords(3,15)
        circ2 = cliffords(3,20)
        t1 = tensorfy(circ1)
        t2 = tensorfy(circ2)
        comp1 = compose_tensors(t1,t2)
        circ1.compose(circ2)
        comp2 = tensorfy(circ1)
        self.assertTrue(compare_tensors(comp1,comp2))

    def test_adjoint(self):
        random.seed(SEED)
        circ = cliffords(3, 16)
        t = tensorfy(circ)
        t_adj = adjoint(t)
        circ_adj = tensorfy(circ.adjoint())
        self.assertTrue(compare_tensors(t_adj,circ_adj))

    def test_multiedge_scalar(self):
        g = Multigraph()
        g.set_auto_simplify(False)
        i1 = g.add_vertex(1,0,0)
        i2 = g.add_vertex(2,1,0)
        g.add_edges([(i1, i2)] * 3)
        self.assertTrue(compare_tensors(g, np.array([np.sqrt(2)**(-1)]), preserve_scalar=True))

    def test_self_loop_scalar(self):
        g = Multigraph()
        g.set_auto_simplify(False)
        i1 = g.add_vertex(1,0,0)
        g.add_edge((i1, i1))
        self.assertTrue(compare_tensors(g, np.array([2]), preserve_scalar=True))
        g.add_edge((i1, i1), 2)
        self.assertTrue(compare_tensors(g, np.array([0]), preserve_scalar=True))

    def test_self_loop_state(self):
        g = Multigraph()
        g.set_auto_simplify(False)
        i0 = g.add_vertex(0,0,0)
        i1 = g.add_vertex(2,0,1)
        g.set_inputs((i0,))
        g.add_edge((i0, i1))
        self.assertTrue(compare_tensors(g, np.array([1,0])))
        g.add_edge((i1, i1), 2)
        self.assertTrue(compare_tensors(g, np.array([0,1])))

    def test_self_loop_and_parallel_edge_map(self):
        g = Multigraph()
        g.set_auto_simplify(False)
        i0 = g.add_vertex(0,0,0)
        i1 = g.add_vertex(2,0,1)
        i2 = g.add_vertex(1,0,2)
        i3 = g.add_vertex(0,0,3)
        g.set_inputs((i0,))
        g.set_outputs((i3,))
        g.add_edges([(i0, i1), (i1, i1)] + [(i1, i2)] * 2)
        g.add_edges([(i2, i2), (i2, i3)], 2)
        self.assertTrue(compare_tensors(g, np.array([[0,0],[1,0]])))

    def test_to_tensor_equivalent(self):
        g = Graph()
        g.add_vertex(VertexType.Z, phase=1)
        g1 = Graph()
        g1.add_vertex(VertexType.X, phase=1)
        self.assertTrue(g.to_tensor() == g1.to_tensor())


if __name__ == '__main__':
    unittest.main()



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from . import Circuit
from .gates import *
from fractions import Fraction

def parse_qsim(data: str) -> Circuit:
    """Produces a :class:`Circuit` based on a .qsim description of a circuit."""
    lines = data.strip().splitlines()

    try:
        qcount = int(lines.pop(0))
    except ValueError:
        raise ValueError('First line should be qubit count')

    gates: List[Gate] = []

    for l in lines:
        l = l.strip()
        if l == '': continue
        gdesc = l.split(' ')
        q = int(gdesc[2])
        if gdesc[1] == 'rz':
            phase = float(gdesc[3]) / math.pi
            gates.append(ZPhase(target = q, phase = Fraction(phase)))
        elif gdesc[1] == 'hz_1_2':
            gates.append(ZPhase(target = q, phase = Fraction(1,4)))
            gates.append(XPhase(target = q, phase = Fraction(1,2)))
            gates.append(ZPhase(target = q, phase = Fraction(-1,4)))
        elif gdesc[1] == 'x_1_2':
            gates.append(XPhase(target = q, phase = Fraction(1,2)))
        elif gdesc[1] == 'y_1_2':
            gates.append(XPhase(target = q, phase = Fraction(1,2)))
            gates.append(ZPhase(target = q, phase = Fraction(1,2)))
            gates.append(XPhase(target = q, phase = Fraction(-1,2)))
        elif gdesc[1] == 'fs':
            q1 = int(gdesc[3])
            theta = float(gdesc[4]) / math.pi
            phi = float(gdesc[5]) / math.pi
            gates.append(FSim(q, q1, Fraction(theta), Fraction(phi)))

    c = Circuit(qcount)
    c.gates = gates
    return c
